import { BlockInventory } from "./block/inventory.js";
import { BlockPermutation } from "./block/permutation.js";
import { BlockTags } from "./block/tags.js";
import { BlockType } from "./block/type.js";
import { Dimension } from "./dimension.js";
import { Location } from "./location.js";
import { ItemStack } from "./item/stack.js";
import { Player } from "./player.js";
import { Wrapper } from "./wrapper.js";
import { Direction, LiquidType } from "@minecraft/server";
import * as I from "./inspect.js";
import * as PP from "./pprint.js";
export { BlockPermutation, BlockTags, BlockType, LiquidType };
export { BlockStates } from "./block/states.js";
export class Block extends Wrapper {
    #dimension;
    #location;
    #tags;
    get dimension() {
        if (!this.#dimension)
            this.#dimension = new Dimension(this.raw.dimension);
        return this.#dimension;
    }
    get inventory() {
        if (this.#inventory === undefined) {
            const raw = this.raw.getComponent(BlockInventory.typeId);
            this.#inventory = raw ? new BlockInventory(raw) : null;
        }
        return this.#inventory ?? undefined;
    }
    #inventory;
    get isAir() {
        return this.raw.isAir;
    }
    get isLiquid() {
        return this.raw.isLiquid;
    }
    get isSolid() {
        return this.raw.isSolid;
    }
    get isValid() {
        return this.raw.isValid;
    }
    get isWaterlogged() {
        return this.raw.isWaterlogged;
    }
    set isWaterlogged(waterlogged) {
        this.raw.setWaterlogged(waterlogged);
    }
    get location() {
        if (!this.#location)
            this.#location = new Location(this.raw.location);
        return this.#location;
    }
    get permutation() {
        // Can't cache the permutation because it might be stale.
        return new BlockPermutation(this.raw.permutation);
    }
    set permutation(newPerm) {
        this.raw.setPermutation(newPerm.raw);
    }
    get redstonePower() {
        return this.raw.getRedstonePower();
    }
    get tags() {
        // We can cache this because BlockTags does not hold a copy.
        if (!this.#tags)
            this.#tags = new BlockTags(this.raw);
        return this.#tags;
    }
    get type() {
        // Can't cache the block type because it might be stale.
        return new BlockType(this.raw.type);
    }
    set type(blockType) {
        this.raw.setType(blockType);
    }
    get typeId() {
        return this.raw.typeId;
    }
    set typeId(typeId) {
        this.raw.setType(typeId);
    }
    get x() {
        return this.raw.x;
    }
    get y() {
        return this.raw.y;
    }
    get z() {
        return this.raw.z;
    }
    canBeDestroyedByLiquidSpread(liquidType) {
        return this.raw.canBeDestroyedByLiquidSpread(liquidType);
    }
    canContainLiquid(liquidType) {
        return this.raw.canContainLiquid(liquidType);
    }
    isLiquidBlocking(liquidType) {
        return this.raw.isLiquidBlocking(liquidType);
    }
    liquidSpreadCausesSpawn(liquidType) {
        return this.raw.liquidSpreadCausesSpawn(liquidType);
    }
    liquidCanFlowFromDirection(liquidType, flowDirection) {
        return this.raw.liquidCanFlowFromDirection(liquidType, flowDirection);
    }
    offset(...args) {
        switch (args.length) {
            case 1:
            case 2:
                const d = args[1] ?? 1;
                switch (args[0]) {
                    case Direction.Down: return this.offset(0, -d, 0);
                    case Direction.East: return this.offset(d, 0, 0);
                    case Direction.North: return this.offset(0, 0, -d);
                    case Direction.South: return this.offset(0, 0, d);
                    case Direction.Up: return this.offset(0, d, 0);
                    case Direction.West: return this.offset(-d, 0, 0);
                    default:
                        throw new TypeError(`Invalid direction: ${args[0]}`);
                }
            case 3:
                return this.dimension.getBlock(this.location.offset(args[0], args[1], args[2]));
            default:
                throw new Error("impossible");
        }
    }
    /** Package private */
    getComponentOrThrow(componentId) {
        const c = this.raw.getComponent(componentId);
        if (c) {
            return c;
        }
        else {
            throw new TypeError(`The block does not have a component \`${componentId}'`);
            ;
        }
    }
    getItemStack(amount, withData) {
        const rawSt = this.raw.getItemStack(amount, withData);
        return rawSt ? new ItemStack(rawSt) : undefined;
    }
    // A workaround for a possible TypeScript bug: Using
    // I.customInspectSymbol here causes mixins to fail typechecking,
    // apparently because it's a unique
    // symbol. https://github.com/microsoft/TypeScript/issues/57165
    /// @internal
    [Symbol.for("cicada-lib.inspect")](inspect, stylise) {
        const obj = {
            dimension: this.dimension,
            location: this.location,
            permutation: this.permutation,
            isAir: this.isAir,
            isLiquid: this.isLiquid,
            isSolid: this.isSolid,
        };
        if (this.canContainLiquid(LiquidType.Water)) {
            obj.isWaterlogged = this.isWaterlogged;
        }
        // THINKME: What about other liquid-related properties? Are they
        // worth reporting? (Yes they are!)
        if (this.redstonePower !== undefined)
            obj.redstonePower = this.redstonePower;
        // NOTE: The class Block doesn't have getComponents() for some
        // reason. We can only inspect components known to us.
        const comps = new Set();
        if (this.inventory)
            comps.add(this.inventory);
        if (comps.size > 0)
            obj.components = comps;
        return PP.spaceCat(stylise(PP.text("Block"), I.TokenType.Class), inspect(obj));
    }
}
function IsBlockEvent(base) {
    class IsBlockEvent extends base {
        get block() {
            return new Block(this.raw.block);
        }
        dimension() {
            return new Dimension(this.raw.dimension);
        }
    }
    return IsBlockEvent;
}
export class PlayerBreakBlockBeforeEvent extends IsBlockEvent((Wrapper)) {
    cancel() {
        this.raw.cancel = true;
    }
    /** The item stack that is being used to break the block, or undefined
     * if empty hand. Note that the returned object is only a copy of the
     * item the player is using. It does not reflect future state changes,
     * and mutating it has no effects on the original stack even after
     * leaving the read-only mode.
     */
    get itemStack() {
        return this.raw.itemStack
            ? new ItemStack(this.raw.itemStack)
            : undefined;
    }
    set itemStack(st) {
        // @ts-ignore: Override exactOptionalPropertyTypes
        this.raw.itemStack = st ? st.raw : undefined;
    }
    get player() {
        return new Player(this.raw.player);
    }
}
//# sourceMappingURL=block.js.map
