/** For internal use only */
export class Chunk {
    #buffer;
    #offset;
    #length;
    #capacity;
    // These are invalidated when #offset or #capacity changes.
    #dView;
    #u8View;
    static allocate(length) {
        return new Chunk(new ArrayBuffer(length), 0, 0);
    }
    static wrap(octets) {
        return new Chunk(octets.buffer, octets.byteOffset, octets.byteLength, octets.byteLength);
    }
    static copyFrom(octets) {
        return Chunk.wrap(new Uint8Array(octets));
    }
    constructor(buffer, offset = 0, length = buffer.byteLength - offset, capacity = buffer.byteLength - offset) {
        this.#buffer = buffer;
        this.#offset = offset;
        this.#length = length;
        this.#capacity = capacity;
        this.#dView = null;
        this.#u8View = null;
    }
    /** Iterate over octets in the chunk. The iterator is only valid until
     * the next time the chunk is mutated.
     */
    *[Symbol.iterator]() {
        for (let i = 0; i < this.#length; i++)
            yield this.u8View[i];
    }
    get length() {
        return this.#length;
    }
    set length(len) {
        if (len >= 0 && len <= this.#capacity) {
            if (this.#length != len) {
                this.#length = len;
            }
        }
        else {
            throw new RangeError(`New length ${len} is out of range`);
        }
    }
    get unused() {
        return this.#capacity - this.#length;
    }
    get dView() {
        if (!this.#dView) {
            this.#dView = new DataView(this.#buffer, this.#offset, this.#capacity);
        }
        return this.#dView;
    }
    get u8View() {
        if (!this.#u8View) {
            this.#u8View = new Uint8Array(this.#buffer, this.#offset, this.#capacity);
        }
        return this.#u8View;
    }
    toUint8Array() {
        if (this.#u8View && this.#capacity == this.#length) {
            return this.#u8View;
        }
        else {
            return new Uint8Array(this.#buffer, this.#offset, this.#length);
        }
    }
    unsafeSubChunk(begin, end) {
        begin ??= 0;
        if (begin < 0) {
            begin += this.#length;
        }
        if (begin > this.#length) {
            begin = this.#length;
        }
        end ??= this.#length;
        if (end < 0) {
            end += this.#length;
        }
        if (end > this.#length) {
            end = this.#length;
        }
        const actualLen = end - begin;
        return new Chunk(this.#buffer, this.#offset + begin, actualLen, actualLen);
    }
    #invalidateViews() {
        this.#dView = null;
        this.#u8View = null;
    }
    clear() {
        this.#length = 0;
    }
    append(octets) {
        this.u8View.set(octets, this.#length);
        this.#length += octets.byteLength;
    }
    take(length) {
        const actualLen = Math.min(length, this.#length);
        const taken = new Chunk(this.#buffer, this.#offset, actualLen, actualLen);
        this.#offset += actualLen;
        this.#capacity -= actualLen;
        this.#length -= actualLen;
        this.#invalidateViews();
        return taken;
    }
    drop(length) {
        const actualLen = Math.min(length, this.#length);
        this.#offset += actualLen;
        this.#capacity -= actualLen;
        this.#length -= actualLen;
        this.#invalidateViews();
    }
}
//# sourceMappingURL=chunk.js.map
