import { Hasher } from "../hasher.js";
import { map } from "../iterable.js";
/** Unordered finite map, similar to the built-in `Map` but can have
 * user-supplied hash function and equality. The type `K` can be anything
 * while the type `V` may not be inhabited by `undefined`. Any attempts on
 * inserting `undefined` as a value will result in a `TypeError`.
 *
 * This implementation does not maintain a hash table on its own. It
 * instead relies on the standard `Map` class internally.
 */
export class HashMap {
    #eq;
    #hash;
    #hasher;
    #buckets;
    // The number of key/value pairs in the map.
    #size;
    constructor(...args) {
        this.#hasher = new Hasher();
        this.#buckets = new Map();
        this.#size = 0;
        switch (args.length) {
            case 0:
                this.#eq = (defaultEq);
                this.#hash = (defaultHash);
                break;
            case 1:
                if (args[0] instanceof HashMap) {
                    // Special case: we can safely clone the map.
                    this.#eq = args[0].#eq;
                    this.#hash = args[0].#hash;
                    this.#buckets = new Map(map(args[0].#buckets, ([hash, bucket]) => [hash, bucket.clone()]));
                    this.#size = args[0].#size;
                }
                else {
                    this.#eq = (defaultEq);
                    this.#hash = (defaultHash);
                    for (const [k, v] of args[0]) {
                        this.set(k, v);
                    }
                }
                break;
            case 2:
                this.#eq = args[0];
                this.#hash = args[1];
                break;
            case 3:
                // We cannot clone the map even if the Iterable is actually
                // a HashMap, because hash functions may not be the same.
                this.#eq = args[1];
                this.#hash = args[2];
                for (const [k, v] of args[0]) {
                    this.set(k, v);
                }
                break;
            default:
                throw new TypeError("Wrong number of arguments");
        }
    }
    /** The number of key/value pairs in the map. */
    get size() {
        return this.#size;
    }
    /** This is identical to {@link entries}. */
    [Symbol.iterator]() {
        return this.entries();
    }
    get [Symbol.toStringTag]() {
        return "HashMap";
    }
    /** Remove all elements from the map. */
    clear() {
        this.#buckets.clear();
        this.#size = 0;
    }
    /** Delete a key and its value from the map. Return `true` iff the key
     * was present.
     */
    "delete"(key) {
        const hash = this.#hashFor(key);
        const bucket = this.#buckets.get(hash);
        if (bucket) {
            if (bucket.delete(key, this.#eq)) {
                if (bucket.isEmpty)
                    this.#buckets.delete(hash);
                this.#size--;
                return true;
            }
        }
        return false;
    }
    /** Delete a single key/value pair in the map and return it if any. Do
     * nothing if the map is empty.
     */
    deleteAny() {
        for (const [hash, bucket] of this.#buckets) {
            const pair = bucket.deleteAny();
            if (pair) {
                if (bucket.isEmpty)
                    this.#buckets.delete(hash);
                this.#size--;
                return pair;
            }
        }
        return undefined;
    }
    /** Iterate over key/value pairs in the map. */
    *entries() {
        for (const bucket of this.#buckets.values()) {
            yield* bucket;
        }
    }
    /** Apply the given function to each key/value pair in the map. */
    forEach(f, thisArg) {
        const boundF = f.bind(thisArg);
        for (const bucket of this.#buckets.values()) {
            for (const [k, v] of bucket) {
                boundF(v, k, this);
            }
        }
    }
    /** Lookup the value at a key in the map, or return `undefined` if no
     * corresponding value exists.
     */
    "get"(key) {
        const hash = this.#hashFor(key);
        const bucket = this.#buckets.get(hash);
        return bucket?.get(key, this.#eq);
    }
    /** See if a key is in the map. */
    has(key) {
        const hash = this.#hashFor(key);
        const bucket = this.#buckets.get(hash);
        return bucket ? bucket.has(key, this.#eq) : false;
    }
    /** Iterate over keys in the map. */
    *keys() {
        for (const bucket of this.#buckets.values()) {
            yield* bucket.keys();
        }
    }
    /** Insert a new key and a value in the map. If the key is already
     * present in the map, the associated value is replaced with the
     * supplied one.
     *
     * If a combining function is supplied, and there is an old value for
     * the same key, the method instead replaces the old value with
     * `combineFn(oldValue, newValue, newKey)`.
     */
    "set"(key, value, combineFn) {
        if (key === undefined)
            throw new TypeError("`undefined' is not a valid key of HashMap");
        const hash = this.#hashFor(key);
        const bucket = this.#buckets.get(hash);
        if (bucket) {
            if (bucket.set(key, value, this.#eq, combineFn))
                this.#size++;
        }
        else {
            this.#buckets.set(hash, new Bucket(key, value));
            this.#size++;
        }
        return this;
    }
    /** Iterate over values in the map. */
    *values() {
        for (const bucket of this.#buckets.values()) {
            yield* bucket.values();
        }
    }
    #hashFor(key) {
        this.#hash(this.#hasher, key);
        const hash = this.#hasher.final();
        this.#hasher.reset();
        return hash;
    }
}
function defaultEq(a, b) {
    return a === b;
}
function defaultHash(hasher, k) {
    hasher.update(k);
}
class Bucket {
    #entries;
    constructor(...args) {
        this.#entries =
            args.length === 0
                ? null
                : { value: args, next: null };
    }
    get isEmpty() {
        return this.#entries == null;
    }
    clone() {
        const ret = new Bucket();
        for (let cell = this.#entries, prev = null; cell; cell = cell.next) {
            const cloned = { value: cell.value, next: null };
            if (prev)
                prev.next = cloned;
            else
                ret.#entries = cloned;
            prev = cloned;
        }
        return ret;
    }
    "get"(key, eq) {
        for (let cell = this.#entries; cell; cell = cell.next) {
            if (eq(cell.value[0], key))
                return cell.value[1];
        }
        return undefined;
    }
    has(key, eq) {
        for (let cell = this.#entries; cell; cell = cell.next) {
            if (eq(cell.value[0], key))
                return true;
        }
        return false;
    }
    /** Return `true` iff the size of the bucket increased. */
    "set"(key, value, eq, combineFn) {
        for (let cell = this.#entries; cell; cell = cell.next) {
            if (eq(cell.value[0], key)) {
                if (combineFn)
                    cell.value[1] = combineFn(cell.value[1], value, cell.value[0]);
                else
                    cell.value[1] = value;
                return false;
            }
        }
        this.#entries = { value: [key, value], next: this.#entries };
        return true;
    }
    /** Return `true` iff the key was present. */
    "delete"(key, eq) {
        for (let cell = this.#entries, prev = null; cell; prev = cell, cell = cell.next) {
            if (eq(cell.value[0], key)) {
                if (prev)
                    prev.next = cell.next;
                else
                    this.#entries = cell.next;
                return true;
            }
        }
        return false;
    }
    deleteAny() {
        if (this.#entries) {
            const pair = this.#entries.value;
            this.#entries = this.#entries.next;
            return pair;
        }
        return undefined;
    }
    *[Symbol.iterator]() {
        for (let cell = this.#entries; cell; cell = cell.next) {
            yield cell.value;
        }
    }
    *keys() {
        for (let cell = this.#entries; cell; cell = cell.next) {
            yield cell.value[0];
        }
    }
    *values() {
        for (let cell = this.#entries; cell; cell = cell.next) {
            yield cell.value[1];
        }
    }
}
//# sourceMappingURL=hash-map.js.map
