import { Hasher } from "../hasher.js";
import { map } from "../iterable.js";
/** Unordered finite set, similar to the built-in `Set` but can have
 * user-supplied hash function and equality. The type `T` may not be
 * inhabited by `undefined`. Any attempts on inserting `undefined` as an
 * element will result in a `TypeError`.
 *
 * This implementation does not maintain a hash table on its own. It
 * instead relies on the standard `Map` class internally.
 */
export class HashSet {
    #eq;
    #hash;
    #hasher;
    #buckets;
    // The number of elements in the set.
    #size;
    constructor(...args) {
        this.#hasher = new Hasher();
        this.#buckets = new Map();
        this.#size = 0;
        switch (args.length) {
            case 0:
                this.#eq = (defaultEq);
                this.#hash = (defaultHash);
                break;
            case 1:
                if (args[0] instanceof HashSet) {
                    // Special case: we can safely clone the set.
                    this.#eq = args[0].#eq;
                    this.#hash = args[0].#hash;
                    this.#buckets = new Map(map(args[0].#buckets, ([hash, bucket]) => [hash, bucket.clone()]));
                    this.#size = args[0].#size;
                }
                else {
                    this.#eq = (defaultEq);
                    this.#hash = (defaultHash);
                    for (const e of args[0]) {
                        this.add(e);
                    }
                }
                break;
            case 2:
                this.#eq = args[0];
                this.#hash = args[1];
                break;
            case 3:
                // We cannot clone the set even if the Iterable is actually
                // a HashSet, because hash functions may not be the same.
                this.#eq = args[1];
                this.#hash = args[2];
                for (const e of args[0]) {
                    this.add(e);
                }
                break;
            default:
                throw new TypeError("Wrong number of arguments");
        }
    }
    /** The number of elements in the set. */
    get size() {
        return this.#size;
    }
    /** This is identical to {@link values}. */
    [Symbol.iterator]() {
        return this.values();
    }
    get [Symbol.toStringTag]() {
        return "HashSet";
    }
    /** Insert a new element in the set. If the element is already
     * present in the set, it is replaced with the supplied one.
     */
    add(element) {
        if (element === undefined)
            throw new TypeError("`undefined' is not a valid element of HashSet");
        const hash = this.#hashFor(element);
        const bucket = this.#buckets.get(hash);
        if (bucket) {
            if (bucket.add(element, this.#eq))
                this.#size++;
        }
        else {
            this.#buckets.set(hash, new Bucket(element));
            this.#size++;
        }
        return this;
    }
    /** Remove all elements from the set. */
    clear() {
        this.#buckets.clear();
        this.#size = 0;
    }
    /** Delete an element from the set. Return `true` iff the element was
     * present.
     */
    "delete"(element) {
        const hash = this.#hashFor(element);
        const bucket = this.#buckets.get(hash);
        if (bucket) {
            if (bucket.delete(element, this.#eq)) {
                if (bucket.isEmpty)
                    this.#buckets.delete(hash);
                this.#size--;
                return true;
            }
        }
        return false;
    }
    /** Delete a single element in the set and return it if any. Do nothing
     * if the set is empty.
     */
    deleteAny() {
        for (const [hash, bucket] of this.#buckets) {
            const elem = bucket.deleteAny();
            if (elem) {
                if (bucket.isEmpty)
                    this.#buckets.delete(hash);
                this.#size--;
                return elem;
            }
        }
        return undefined;
    }
    /** Return a new set consisting of elements of `this` not existing in
     * the set `s`.
     */
    difference(s) {
        // NOTE: If two sets have an identical equality and a hash
        // function, maybe we can optimize this a bit by directly working
        // on buckets from the two sets. I don't think it's worth it
        // though.
        const ret = new HashSet(this.#eq, this.#hash);
        for (const elem of this.values()) {
            if (!s.has(elem))
                ret.add(elem);
        }
        return ret;
    }
    /** Iterate over elements in the set. */
    *entries() {
        for (const bucket of this.#buckets.values()) {
            for (const elem of bucket) {
                yield [elem, elem];
            }
        }
    }
    /** Apply the given function to each element in the set. */
    forEach(f, thisArg) {
        const boundF = f.bind(thisArg);
        for (const bucket of this.#buckets.values()) {
            for (const elem of bucket) {
                boundF(elem, elem, this);
            }
        }
    }
    /** See if an element is in the set. */
    has(element) {
        const hash = this.#hashFor(element);
        const bucket = this.#buckets.get(hash);
        return bucket ? bucket.has(element, this.#eq) : false;
    }
    /** Return a new set consisting of elements of `this` that are also
     * existing in the set `s`.
     */
    intersection(s) {
        // See a comment in difference().
        const ret = new HashSet(this.#eq, this.#hash);
        for (const elem of this.values()) {
            if (s.has(elem))
                ret.add(elem);
        }
        return ret;
    }
    /** Return `true` iff `this` and the set `s` contain no common
     * elements.
     */
    isDisjointFrom(s) {
        // See a comment in difference().
        for (const elem of this.values()) {
            if (s.has(elem))
                return false;
        }
        return true;
    }
    /** Return `true` iff every element of `this` is also an element of
     * `s`.
     */
    isSubsetOf(s) {
        // See a comment in difference().
        for (const elem of this.values()) {
            if (!s.has(elem))
                return false;
        }
        return true;
    }
    /** Return `true` iff every element of `s` is also an element of
     * `this`.
     */
    isSupersetOf(s) {
        // See a comment in difference().
        for (const elem of s.values()) {
            if (!this.has(elem))
                return false;
        }
        return true;
    }
    /** Iterate over elements in the set. This is identical to {@link
     * values}.
     */
    keys() {
        return this.values();
    }
    /** Return a new set consisting of elements of either `this` or `s` but
     * not both.
     */
    symmetricDifference(s) {
        // See a comment in difference().
        const s1 = this.difference(s);
        const s2 = s.difference(this);
        return s1.union(s2);
    }
    /** Take the left-biased union of `this` and `s`. The function returns
     * a new set without modifying existing ones.
     */
    union(s) {
        // See a comment in difference().
        const ret = new HashSet(this);
        for (const elem of s.values()) {
            if (!ret.has(elem)) {
                ret.add(elem);
            }
        }
        return ret;
    }
    /** Iterate over elements in the set. */
    *values() {
        for (const bucket of this.#buckets.values()) {
            yield* bucket;
        }
    }
    #hashFor(element) {
        this.#hash(this.#hasher, element);
        const hash = this.#hasher.final();
        this.#hasher.reset();
        return hash;
    }
}
function defaultEq(a, b) {
    return a === b;
}
function defaultHash(hasher, k) {
    hasher.update(k);
}
class Bucket {
    #entries;
    constructor(...args) {
        this.#entries =
            args.length === 0
                ? null
                : { value: args[0], next: null };
    }
    get isEmpty() {
        return this.#entries == null;
    }
    /** Return `true` iff the size of the bucket increased. */
    "add"(elem, eq) {
        for (let cell = this.#entries; cell; cell = cell.next) {
            if (eq(cell.value, elem))
                return false;
        }
        this.#entries = { value: elem, next: this.#entries };
        return true;
    }
    clone() {
        const ret = new Bucket();
        // Can't just copy this.#entries because it's mutable (and we
        // actually mutate it).
        for (let cell = this.#entries, prev = null; cell; cell = cell.next) {
            const cloned = { value: cell.value, next: null };
            if (prev)
                prev.next = cloned;
            else
                ret.#entries = cloned;
            prev = cloned;
        }
        return ret;
    }
    has(elem, eq) {
        for (let cell = this.#entries; cell; cell = cell.next) {
            if (eq(cell.value, elem))
                return true;
        }
        return false;
    }
    /** Return `true` iff the element was present. */
    "delete"(elem, eq) {
        for (let cell = this.#entries, prev = null; cell; prev = cell, cell = cell.next) {
            if (eq(cell.value, elem)) {
                if (prev)
                    prev.next = cell.next;
                else
                    this.#entries = cell.next;
                return true;
            }
        }
        return false;
    }
    deleteAny() {
        if (this.#entries) {
            const elem = this.#entries.value;
            this.#entries = this.#entries.next;
            return elem;
        }
        return undefined;
    }
    *[Symbol.iterator]() {
        for (let cell = this.#entries; cell; cell = cell.next) {
            yield cell.value;
        }
    }
}
//# sourceMappingURL=hash-set.js.map
