import { lazy } from "../lazy.js";
const SHaveOutput = Symbol("SHaveOutput");
const SNeedInput = Symbol("SNeedInput");
const SLeftover = Symbol("SLeftover");
const SHaveInput = Symbol("SHaveInput");
const SClosed = Symbol("SClosed");
export class Conduit {
    async *[Symbol.asyncIterator]() {
        return yield* this[Symbol.iterator]();
    }
    fuse(down) {
        const up = this;
        class Fused extends Conduit {
            *[Symbol.iterator]() {
                let upG;
                const downG = down[Symbol.iterator]();
                let upSup;
                let downSup;
                let goDown = true;
                const leftovers = [];
                while (true) {
                    if (goDown) {
                        const ret = downSup ? downG.next(downSup) : downG.next();
                        downSup = undefined;
                        if (ret.done) {
                            // The downstream is now closed, which means we
                            // no longer run the upstream also.
                            return ret.value;
                        }
                        else {
                            const step = ret.value;
                            switch (step.type) {
                                case SHaveOutput:
                                    // The downstream yielded an output,
                                    // which means this fused conduit
                                    // should also yield it.
                                    yield step;
                                    break;
                                case SNeedInput:
                                    // The downstream needs an input. Do we
                                    // have any leftovers?
                                    if (leftovers.length > 0) {
                                        // Yup. Give it back to the
                                        // downstream.
                                        downSup = {
                                            type: SHaveInput,
                                            input: leftovers.pop()
                                        };
                                    }
                                    else {
                                        // Nope. Switch to the upstream.
                                        goDown = false;
                                    }
                                    break;
                                case SLeftover:
                                    // The downstream yielded a leftover input,
                                    // so the next time it requests for an
                                    // input give the leftover back to the
                                    // downstream.
                                    leftovers.push(step.leftover);
                                    break;
                            }
                        }
                    }
                    else { // !goDown
                        if (!upG)
                            // We haven't even started the upstream. Do it
                            // now.
                            upG = up[Symbol.iterator]();
                        const ret = upSup ? upG.next(upSup) : upG.next();
                        upSup = undefined;
                        if (ret.done) {
                            // Tell the downstream, which is awaiting for
                            // an input, that the upstream is now closed.
                            downSup = {
                                type: SClosed,
                                result: ret.value
                            };
                            goDown = true;
                        }
                        else {
                            const step = ret.value;
                            switch (step.type) {
                                case SHaveOutput:
                                    // The upstream yielded an output. Feed
                                    // the downstream with it.
                                    downSup = {
                                        type: SHaveInput,
                                        input: step.output
                                    };
                                    goDown = true;
                                    break;
                                case SNeedInput:
                                    // Ask the upstream's upstream for an input
                                    // so that we can resume the upstream with
                                    // it.
                                    upSup = yield step;
                                    break;
                                case SLeftover:
                                    // The upstream yielded a leftover
                                    // input, so the fused conduit should
                                    // also yield a leftover.
                                    yield step;
                                    break;
                            }
                        }
                    }
                }
            }
            async *[Symbol.asyncIterator]() {
                // This is mostly the same as
                // Fused.prototype[@@iterator]()... We hate code
                // duplication but we don't think it's avoidable.
                let upG;
                const downG = down[Symbol.asyncIterator]();
                let upSup;
                let downSup;
                let goDown = true;
                const leftovers = [];
                while (true) {
                    if (goDown) {
                        const ret = await (downSup ? downG.next(downSup) : downG.next());
                        downSup = undefined;
                        if (ret.done) {
                            // The downstream is now closed, which means we
                            // no longer run the upstream also.
                            return ret.value;
                        }
                        else {
                            const step = ret.value;
                            switch (step.type) {
                                case SHaveOutput:
                                    // The downstream yielded an output,
                                    // which means this fused conduit
                                    // should also yield it.
                                    yield step;
                                    break;
                                case SNeedInput:
                                    // The downstream needs an input. Do we
                                    // have any leftovers?
                                    if (leftovers.length > 0) {
                                        // Yup. Give it back to the
                                        // downstream.
                                        downSup = {
                                            type: SHaveInput,
                                            input: leftovers.pop()
                                        };
                                    }
                                    else {
                                        // Nope. Switch to the upstream.
                                        goDown = false;
                                    }
                                    break;
                                case SLeftover:
                                    // The downstream yielded a leftover input,
                                    // so the next time it requests for an
                                    // input give the leftover back to the
                                    // downstream.
                                    leftovers.push(step.leftover);
                                    break;
                            }
                        }
                    }
                    else { // !goDown
                        if (!upG)
                            // We haven't even started the upstream. Do it
                            // now.
                            upG = up[Symbol.asyncIterator]();
                        const ret = await (upSup ? upG.next(upSup) : upG.next());
                        upSup = undefined;
                        if (ret.done) {
                            // Tell the downstream, which is awaiting for
                            // an input, that the upstream is now closed.
                            downSup = {
                                type: SClosed,
                                result: ret.value
                            };
                            goDown = true;
                        }
                        else {
                            const step = ret.value;
                            switch (step.type) {
                                case SHaveOutput:
                                    // The upstream yielded an output. Feed
                                    // the downstream with it.
                                    downSup = {
                                        type: SHaveInput,
                                        input: step.output
                                    };
                                    goDown = true;
                                    break;
                                case SNeedInput:
                                    // Ask the upstream's upstream for an input
                                    // so that we can resume the upstream with
                                    // it.
                                    upSup = yield step;
                                    break;
                                case SLeftover:
                                    // The upstream yielded a leftover
                                    // input, so the fused conduit should
                                    // also yield a leftover.
                                    yield step;
                                    break;
                            }
                        }
                    }
                }
            }
        }
        return new Fused();
    }
    run() {
        const ret = this[Symbol.iterator]().next();
        if (ret.done)
            return ret.value;
        else
            throw new TypeError(`The conduit yielded when it shouldn't: ${ret.value}`);
    }
    async runAsync() {
        const ret = await this[Symbol.asyncIterator]().next();
        if (ret.done)
            return ret.value;
        else
            throw new TypeError(`The conduit yielded when it shouldn't: ${ret.value}`);
    }
}
// Primitives
export function conduit(f) {
    class Wrapped extends Conduit {
        [Symbol.iterator]() {
            return f();
        }
        [Symbol.for("nodejs.util.inspect.custom")](_depth, _opts, _inspect) {
            return `Wrapped ${f.toString()}`;
        }
    }
    return new Wrapped();
}
export const awaitC = lazy(() => {
    class Await extends Conduit {
        *[Symbol.iterator]() {
            const supply = yield { type: SNeedInput };
            switch (supply.type) {
                case SHaveInput:
                    return supply.input;
                case SClosed:
                    return undefined;
            }
        }
    }
    return new Await();
});
export function awaitForever(f) {
    class AwaitForever extends Conduit {
        *[Symbol.iterator]() {
            while (true) {
                const supply = yield { type: SNeedInput };
                switch (supply.type) {
                    case SHaveInput:
                        yield* f(supply.input);
                        break;
                    case SClosed:
                        return;
                }
            }
        }
    }
    return new AwaitForever();
}
/** Apply an action to all values in the stream. This is a variant of
 * `awaitForever()` where the returned values from the supplied function
 * are ignored.
 */
export function forEach(f) {
    class ForEach extends Conduit {
        *[Symbol.iterator]() {
            while (true) {
                const supply = yield { type: SNeedInput };
                switch (supply.type) {
                    case SHaveInput:
                        f(supply.input);
                        break;
                    case SClosed:
                        return;
                }
            }
        }
    }
    return new ForEach();
}
export function leftover(input) {
    class Leftover extends Conduit {
        *[Symbol.iterator]() {
            yield {
                type: SLeftover,
                leftover: input
            };
        }
    }
    return new Leftover();
}
export function yieldC(output) {
    class Yield extends Conduit {
        *[Symbol.iterator]() {
            yield {
                type: SHaveOutput,
                output
            };
        }
    }
    return new Yield();
}
export class PrematureEOF extends Error {
    constructor(...args) {
        super(...args);
    }
}
//# sourceMappingURL=primitives.js.map
