import { formatWithOptions, stringify } from "./format.js";
import { inspect } from "./inspect.js";
import { Timer } from "./timer.js";
import * as PP from "./pprint.js";
export var Severity;
(function (Severity) {
    Severity[Severity["Debug"] = 0] = "Debug";
    Severity[Severity["Info"] = 1] = "Info";
    Severity[Severity["Log"] = 2] = "Log";
    Severity[Severity["Warning"] = 3] = "Warning";
    Severity[Severity["Error"] = 4] = "Error";
})(Severity || (Severity = {}));
/** Console API for non-interactive wall of text. Intended only for
 * internal use. */
export class ConsoleBase {
    /** A non-standard property that controls the threshold of severity.
     */
    logLevel = Severity.Debug;
    /** A non-standard property for indenting log groups and pretty-printed
     * objects. */
    indentationWidth = 4;
    /** A non-standard property for choosing whether to color the output.
     */
    colors = true;
    // Used for console.count().
    #counts = new Map();
    // Used for console.group().
    #currentGroupDepth = 0;
    // Used for console.time().
    #timers = new Map();
    get #inspectOpts() {
        return {
            indentationWidth: this.indentationWidth,
            colors: this.colors
        };
    }
    assert(assertion, ...args) {
        if (!assertion) {
            /* We think it's better to show a stacktrace, not just a
             * message. */
            this.error(args.length > 0
                ? new Error("Assertion failed: " + this.format(...args))
                : new Error("Assertion failed"));
        }
    }
    clear() {
        // We can't clear the log from within the API. This has to be a
        // no-op.
    }
    count(label = "default") {
        const count = (this.#counts.get(label) || 0) + 1;
        this.debug(`${label}: ${count}`);
        this.#counts.set(label, count);
    }
    countReset(label = "default") {
        this.#counts.delete(label);
    }
    debug(...args) {
        this.logImpl(Severity.Debug, ...args);
    }
    dir(obj) {
        // We can't actually make the result interactive, but at least we
        // can remove limitations and apply formatting codes.
        this.log(inspect(obj, { indentationWidth: this.indentationWidth,
            // THINKME: showHidden makes it overflow
            // the stack often.
            showHidden: false,
            depth: Infinity,
            colors: true,
            maxArrayLength: Infinity,
            maxStringLength: Infinity,
            getters: true }));
    }
    dirxml(obj) {
        // FIXME: What should we do about this? We don't think we have XML
        // DOM in our environment.
        this.dir(obj);
    }
    error(...args) {
        this.logImpl(Severity.Error, ...args);
    }
    group(label) {
        if (label != null) {
            this.log(label);
        }
        this.#currentGroupDepth++;
    }
    groupCollapsed(label) {
        this.group(label); // We can't collapse anything of course.
    }
    groupEnd() {
        if (this.#currentGroupDepth > 0) {
            this.#currentGroupDepth--;
        }
    }
    info(...args) {
        this.logImpl(Severity.Info, ...args);
    }
    log(...args) {
        this.logImpl(Severity.Log, ...args);
    }
    table(obj, _columns) {
        // It is way too hard to do this properly on CLI. Just redirect to
        // .log() and hope nobody will rely on this.
        this.log(obj);
    }
    time(label = "default") {
        if (this.#timers.has(label)) {
            this.warn(`${label}: timer restarted`);
        }
        this.#timers.set(label, new Timer());
    }
    timeLog(label = "default") {
        const timer = this.#timers.get(label);
        if (timer) {
            this.log(`${label}: ${timer}`);
        }
        else {
            this.error(`${label}: timer does not exist`);
        }
    }
    timeEnd(label = "default") {
        const timer = this.#timers.get(label);
        if (timer) {
            this.#timers.delete(label);
            this.log(`${label}: ${timer} - timer ended`);
        }
        else {
            this.error(`${label}: timer does not exist`);
        }
    }
    timeStamp(_label) {
        // No-op. There is no such thing as performance tool on Bedrock.
    }
    trace(...args) {
        const msg = args.length > 0 ? this.format(...args) : "";
        const err = new Error(msg);
        err.name = "Trace";
        this.log(err);
    }
    warn(...args) {
        this.logImpl(Severity.Warning, ...args);
    }
    indent() {
        return " ".repeat(this.#currentGroupDepth * this.indentationWidth);
    }
    #render(doc) {
        return PP.displayS(PP.renderPretty(1.0, Infinity, this.colors ? doc : PP.plain(doc)));
    }
    severityToString(sev) {
        return this.#render(ConsoleBase.#severityToDoc(sev));
    }
    static #severityToDoc(sev) {
        switch (sev) {
            case Severity.Debug: return PP.darkGray(PP.text("[Debug]"));
            case Severity.Info: return PP.gray(PP.text("[Info]"));
            case Severity.Log: return PP.bold(PP.text("[Log]"));
            case Severity.Warning: return PP.yellow(PP.text("[Warning]"));
            case Severity.Error: return PP.bold(PP.red(PP.text("[Error]")));
            default:
                throw Error(`Unknown severity: ${sev}`);
        }
    }
    timestamp() {
        return this.#render(PP.gray(PP.text(`[${new Date().toLocaleTimeString()}]`)));
    }
    format(...args) {
        if (args.length > 0) {
            if (typeof args[0] === "string") {
                /* The first argument being a string means that it may
                 * contain substitution specifications. */
                return formatWithOptions(this.#inspectOpts, args[0], ...(args.slice(1)));
            }
            else {
                return args.map(val => stringify(val, this.#inspectOpts)).join(" ");
            }
        }
        else {
            return "";
        }
    }
}
//# sourceMappingURL=console-base.js.map
