import { conduit, sinkString, takeE, yieldC } from "./conduit.js";
/** The maximum number of octets that a string property can have. It was
 * ~10 KiB before 1.2.10, and was expanded to 128 KiB for entities and 1
 * MiB for worlds, and then narrowed down to 32 KiB on 1.20.40.
 */
export const MAX_STRING_PROPERTY_LENGTH = 32767;
/** A mixin for objects that have dynamic properties, such as Entity or
 * World.
 */
export function HasDynamicProperties(base) {
    class HasDynamicProperties extends base {
        get dynamicPropertyIds() {
            return this.raw.getDynamicPropertyIds();
        }
        get dynamicPropertyTotalByteCount() {
            return this.raw.getDynamicPropertyTotalByteCount();
        }
        clearDynamicProperties() {
            this.raw.clearDynamicProperties();
        }
        getDynamicProperty(identifier, ty) {
            const prop = this.raw.getDynamicProperty(identifier);
            if (ty == null) {
                // No constraints are requested on the type of the property.
                return prop;
            }
            else {
                const expectedType = ty.substring(0, ty.length - 1);
                const actualType = typeof prop;
                if (prop === undefined) {
                    if (ty.endsWith("?")) {
                        // It can also be undefined.
                        return prop;
                    }
                    else {
                        throw new TypeError(`Dynamic property \`${identifier}' is expected to be a ${expectedType} but is actually undefined`);
                    }
                }
                else if (expectedType === "Vector3" && actualType === "object") {
                    return prop;
                }
                else if (expectedType === actualType) {
                    return prop;
                }
                else {
                    throw new TypeError(`Dynamic property \`${identifier} is expected to be a ${expectedType} but is actually a ${actualType}`);
                }
            }
        }
        setDynamicProperty(identifier, value) {
            this.raw.setDynamicProperty(identifier, value);
        }
    }
    return HasDynamicProperties;
}
export function sourceDynamicProperty(src, numChunks, genId) {
    return conduit(function* () {
        for (let i = 0; i < numChunks; i++) {
            const chunk = src.getDynamicProperty(genId(i), "string");
            yield* yieldC(chunk);
        }
    });
}
export function sinkDynamicProperty(dest, oldNumChunks, genId) {
    return conduit(function* () {
        let numChunks = 0;
        for (;; numChunks++) {
            const chunk = yield* takeE(MAX_STRING_PROPERTY_LENGTH).fuse(sinkString);
            if (chunk.length > 0)
                dest.setDynamicProperty(genId(numChunks), chunk);
            else
                break;
        }
        for (let i = numChunks; i < oldNumChunks; i++) {
            // These parts no longer exist. Remove them.
            dest.setDynamicProperty(genId(i), undefined);
        }
        return numChunks;
    });
}
//# sourceMappingURL=dynamic-props.js.map
