import { Block } from "./block.js";
import { Dimension } from "./dimension.js";
import { HasDynamicProperties } from "./dynamic-props.js";
import { EntityBreathable } from "./entity/breathable.js";
import { EntityHealth, EntityLavaMovement, EntityMovement, EntityUnderwaterMovement } from "./entity/attributes.js";
import { EntityEquipment } from "./entity/equipment.js";
import { EntityInventory } from "./entity/inventory.js";
import { EntityRideable } from "./entity/rideable.js";
import { EntityTags } from "./entity/tags.js";
import { lazy } from "./lazy.js";
import { Location } from "./location.js";
import { Wrapper } from "./wrapper.js";
export { EntityBreathable, EntityEquipment, EntityHealth, EntityInventory, EntityLavaMovement, EntityMovement, EntityRideable, EntityTags, EntityUnderwaterMovement, };
export { EquipmentSlot } from "./entity/equipment.js";
export class Entity extends HasDynamicProperties((Wrapper)) {
    get dimension() {
        return new Dimension(this.raw.dimension);
    }
    get id() {
        return this.raw.id;
    }
    get isSneaking() {
        return this.raw.isSneaking;
    }
    get isValid() {
        return this.raw.isValid;
    }
    get location() {
        return new Location(this.raw.location);
    }
    get typeId() {
        return this.raw.typeId;
    }
    /** The set of tags for this entity. The set isn't a snapshot. You can
     * add or remove tags through the standard Set API.
     */
    tags = lazy(() => new EntityTags(this.raw));
    getBlockFromViewDirection(options) {
        let rawHit = this.raw.getBlockFromViewDirection(options);
        return rawHit
            ? {
                block: new Block(rawHit.block),
                face: rawHit.face,
                faceLocation: rawHit.faceLocation
            }
            : undefined;
    }
    kill() {
        this.raw.kill();
    }
    matches(options) {
        return this.raw.matches(options);
    }
    teleport(location, opts) {
        let rawOpts = {};
        if (opts) {
            if (opts.checkForBlocks) {
                rawOpts.checkForBlocks = opts.checkForBlocks;
            }
            if (opts.dimension) {
                rawOpts.dimension = opts.dimension.raw;
            }
            if (opts.facingLocation) {
                rawOpts.facingLocation = opts.facingLocation;
            }
            if (opts.keepVelocity) {
                rawOpts.keepVelocity = opts.keepVelocity;
            }
            if (opts.rotation) {
                rawOpts.rotation = opts.rotation;
            }
        }
        this.raw.teleport(location, rawOpts);
    }
    triggerEvent(eventName) {
        this.raw.triggerEvent(eventName);
    }
    // --------- Components ---------
    get breathable() {
        if (this.#breathable === undefined) {
            const raw = this.raw.getComponent(EntityBreathable.typeId);
            this.#breathable = raw ? new EntityBreathable(raw) : null;
        }
        return this.#breathable ?? undefined;
    }
    #breathable;
    get canClimb() {
        if (this.#canClimb === undefined) {
            this.#canClimb = !!this.raw.getComponent("minecraft:can_climb");
        }
        return this.#canClimb;
    }
    #canClimb;
    get health() {
        if (this.#health === undefined) {
            const raw = this.raw.getComponent(EntityHealth.typeId);
            this.#health = raw ? new EntityHealth(raw) : null;
        }
        return this.#health ?? undefined;
    }
    #health;
    get equipment() {
        if (this.#equipment === undefined) {
            const raw = this.raw.getComponent(EntityEquipment.typeId);
            this.#equipment = raw ? new EntityEquipment(raw) : null;
        }
        return this.#equipment ?? undefined;
    }
    #equipment;
    get inventory() {
        if (this.#inventory === undefined) {
            const raw = this.raw.getComponent(EntityInventory.typeId);
            this.#inventory = raw ? new EntityInventory(raw) : null;
        }
        return this.#inventory ?? undefined;
    }
    #inventory;
    get isHiddenWhenInvisible() {
        if (this.#isHiddenWhenInvisible === undefined) {
            this.#isHiddenWhenInvisible = !!this.raw.getComponent("minecraft:is_hidden_when_invisible");
        }
        return this.#isHiddenWhenInvisible;
    }
    #isHiddenWhenInvisible;
    get lavaMovement() {
        if (this.#lavaMovement === undefined) {
            const raw = this.raw.getComponent(EntityLavaMovement.typeId);
            this.#lavaMovement = raw ? new EntityLavaMovement(raw) : null;
        }
        return this.#lavaMovement ?? undefined;
    }
    #lavaMovement;
    get movement() {
        if (this.#movement === undefined) {
            const raw = this.raw.getComponent(EntityMovement.typeId);
            this.#movement = raw ? new EntityMovement(raw) : null;
        }
        return this.#movement ?? undefined;
    }
    #movement;
    get rideable() {
        if (this.#rideable === undefined) {
            const raw = this.raw.getComponent(EntityRideable.typeId);
            this.#rideable = raw ? new EntityRideable(raw) : null;
        }
        return this.#rideable ?? undefined;
    }
    #rideable;
    get underwaterMovement() {
        if (this.#underwaterMovement === undefined) {
            const raw = this.raw.getComponent(EntityUnderwaterMovement.typeId);
            this.#underwaterMovement = raw ? new EntityUnderwaterMovement(raw) : null;
        }
        ;
        return this.#underwaterMovement ?? undefined;
    }
    #underwaterMovement;
}
/** Package private: user code should not use this. */
export function entityEventOptionsToRaw(opts) {
    let ret = {};
    if (opts.entities) {
        ret.entities = opts.entities.map(e => e.raw);
    }
    if (opts.entityTypes) {
        ret.entityTypes = opts.entityTypes;
    }
    return ret;
}
//# sourceMappingURL=entity.js.map
