/** Package private */
export class CustomEventSignal {
    #handlers;
    constructor() {
        this.#handlers = new Set();
    }
    subscribe(cb) {
        this.#handlers.add(cb);
        return cb;
    }
    unsubscribe(cb) {
        this.#handlers.delete(cb);
    }
    signal(ev) {
        for (const cb of this.#handlers) {
            /* QuickJS, the current interpreter used by Minecraft BE,
             * doesn't provide a mechanism to catch unhandled rejection of
             * promises. But it's common that the listener is actually an
             * async function and it doesn't catch exceptions. When that
             * happens exceptions get lost and cause a great confusion, so
             * if the function returns a promise attach a handler to catch
             * them all.
             *
             * THINKME: We think we can now remove this as of v1.21.70.
             */
            const ret = cb(ev);
            Promise.resolve(ret).catch(e => console.error(e));
        }
    }
}
/** Package private */
export class GluedEventSignalWithoutOptions {
    #signal;
    #glueEv;
    #handlers;
    constructor(rawSignal, glueEv) {
        this.#signal = rawSignal;
        this.#glueEv = glueEv;
        this.#handlers = new Map();
    }
    subscribe(cb) {
        const handler = (rawEv) => {
            /* QuickJS, the current interpreter used by Minecraft BE,
             * doesn't provide a mechanism to catch unhandled rejection of
             * promises. But it's common that the listener is actually an
             * async function and it doesn't catch exceptions. When that
             * happens exceptions get lost and cause a great confusion, so
             * if the function returns a promise attach a handler to catch
             * them all. */
            const ret = cb(this.#glueEv(rawEv));
            Promise.resolve(ret).catch(e => console.error(e));
        };
        this.#signal.subscribe(handler);
        this.#handlers.set(cb, handler);
        return cb;
    }
    unsubscribe(cb) {
        const handler = this.#handlers.get(cb);
        if (handler) {
            this.#signal.unsubscribe(handler);
            this.#handlers.delete(cb);
        }
    }
}
/** Package private */
export class GluedEventSignalWithOptions {
    #signal;
    #glueEv;
    #glueOpts;
    #handlers;
    constructor(rawSignal, glueEv, glueOpts) {
        this.#signal = rawSignal;
        this.#glueEv = glueEv;
        this.#glueOpts = glueOpts;
        this.#handlers = new Map();
    }
    subscribe(cb, opts) {
        const handler = (rawEv) => {
            // See comments on
            // GluedEventSignalWithoutOptions.prototype.subscribe
            const ret = cb(this.#glueEv(rawEv));
            Promise.resolve(ret).catch(e => console.error(e));
        };
        this.#signal.subscribe(handler, opts ? this.#glueOpts(opts) : undefined);
        this.#handlers.set(cb, handler);
        return cb;
    }
    unsubscribe(cb) {
        const handler = this.#handlers.get(cb);
        if (handler) {
            this.#signal.unsubscribe(handler);
            this.#handlers.delete(cb);
        }
    }
}
/** Package private: The sole reason this class exists is that we want to
 * catch unhandled rejection of promises. */
export class PassThruEventSignal {
    #glued;
    constructor(rawSignal) {
        this.#glued = new GluedEventSignalWithoutOptions(rawSignal, (ev) => ev);
    }
    subscribe(cb) {
        this.#glued.subscribe(cb);
        return cb;
    }
    unsubscribe(cb) {
        this.#glued.unsubscribe(cb);
    }
}
//# sourceMappingURL=event.js.map
