import { getProxiedProperty, setProxiedProperty } from "./utils.js";
function clamp(n, min, max, negOffset = max) {
    if (n < 0)
        n += negOffset;
    if (n < min)
        n = min;
    else if (n > max)
        n = max;
    return n;
}
function toArrayIndex(key) {
    switch (typeof key) {
        case "symbol":
            return undefined;
        case "number":
            {
                const uint = key >>> 0;
                if (key === uint && uint !== 0xFFFFFFFF)
                    return uint;
                else
                    return undefined;
            }
        default:
            {
                const uint = Number(key) >>> 0;
                const str = String(uint);
                if (key === str && uint !== 0xFFFFFFFF)
                    return uint;
                else
                    return undefined;
            }
    }
}
/** A base class behaving like a fixed-length sparse array.
 */
export class FixedSparseArrayLike {
    #ops;
    #proxy;
    length;
    constructor(length, ops) {
        this.#ops = {
            has(index) {
                return ops.get(index) !== undefined;
            },
            clone(value) {
                return value;
            },
            ...ops
        };
        this.length = length;
        // The dark side of JavaScript magic... We return an object from
        // the constructor that looks like an array but is actually
        // not. This is necessary to override the behaviour of obj[idx]
        // notation.
        this.#proxy = new Proxy(this, {
            defineProperty(self, key, desc) {
                if (toArrayIndex(key) !== undefined) {
                    // We are unwilling to allow this although it's not
                    // impossible, because we would need to do more work in
                    // "get" and "set" handlers.
                    return false;
                }
                else {
                    Object.defineProperty(self, key, desc);
                    return true;
                }
            },
            deleteProperty(self, key) {
                const index = toArrayIndex(key);
                if (index === undefined) {
                    return false; // Can't delete non-index properties.
                }
                else {
                    self.#ops.delete(index);
                    return true;
                }
            },
            "get"(self, key, receiver) {
                const index = toArrayIndex(key);
                if (index === undefined)
                    return getProxiedProperty(self, key, receiver, FixedSparseArrayLike);
                else
                    return self.#ops.get(index);
            },
            getOwnPropertyDescriptor(self, key) {
                const index = toArrayIndex(key);
                if (index === undefined) {
                    return Object.getOwnPropertyDescriptor(self, key);
                }
                else {
                    const value = self.#ops.get(index);
                    if (value !== undefined)
                        return {
                            value,
                            writable: true,
                            configurable: true,
                            enumerable: true
                        };
                    else
                        return undefined;
                }
            },
            has(self, key) {
                const index = toArrayIndex(key);
                if (index === undefined)
                    return key in self;
                else
                    return self.#ops.has(index);
            },
            ownKeys(self) {
                const keys = Reflect.ownKeys(self);
                for (const index of self.keys()) {
                    keys.push(String(index));
                }
                return keys;
            },
            "set"(self, key, value, receiver) {
                const index = toArrayIndex(key);
                if (index === undefined)
                    setProxiedProperty(self, key, value, receiver, FixedSparseArrayLike);
                else
                    self.#ops.set(index, value);
                return true;
            }
        });
        return this.#proxy;
    }
    // CAUTION: Methods in this very class (but not subclasses) cannot
    // refer to array elements using the this[i] notation due to the way
    // how the magic works. Within methods "this" is bound to a regular
    // object but not a proxy in order to allow accessing private
    // properties, and trying to get its (phantom) numeric properties just
    // results in "undefined". By extension methods in this class cannot do
    // things like Array.prototype.foo.call(this), although it's fine to do
    // Array.prototype.foo.call(this.#proxy). See also
    // getProxiedProperty().
    [Symbol.isConcatSpreadable] = true;
    *[Symbol.iterator]() {
        for (let i = 0; i < this.length; i++) {
            const value = this.#ops.get(i);
            if (value !== undefined) {
                yield value;
            }
        }
    }
    at(index) {
        index = clamp(index, 0, this.length);
        return this.#ops.get(index);
    }
    copyWithin(target, start, end) {
        target = clamp(target, 0, this.length);
        start = clamp(start, 0, this.length);
        if (end !== undefined)
            end = clamp(end, 0, this.length);
        else
            end = this.length;
        const count = Math.min(end - start, this.length - target);
        const dir = (start < target && target < start + count) ? -1 : +1;
        for (let i = 0; i < count;) {
            let from, to;
            if (dir < 0) {
                from = start + count - i - 1;
                to = target + count - i - 1;
            }
            else {
                from = start + i;
                to = target + i;
            }
            const value = this.#ops.get(from);
            if (value === undefined)
                this.#ops.delete(to);
            else
                this.#ops.set(to, this.#ops.clone(value));
        }
        return this;
    }
    concat(...items) {
        return Array.prototype.concat.apply(this.#proxy, items);
    }
    *entries() {
        for (let i = 0; i < this.length; i++) {
            const value = this.#ops.get(i);
            if (value !== undefined) {
                yield [i, value];
            }
        }
    }
    every(p, thisArg) {
        return Array.prototype.every.call(this.#proxy, p, thisArg);
    }
    fill(value, start, end) {
        start = clamp(start, 0, this.length);
        if (end !== undefined)
            end = clamp(end, 0, this.length);
        else
            end = this.length;
        for (let i = start; i < end; i++)
            this.#ops.set(i, this.#ops.clone(value));
        return this;
    }
    filter(p, thisArg) {
        return Array.prototype.filter.call(this.#proxy, p, thisArg);
    }
    find(p, thisArg) {
        // We can't use Array.prototype.find() because it doesn't skip empty slots.
        const boundP = p.bind(thisArg);
        for (const [index, value] of this.entries()) {
            if (!boundP(value, index, this))
                return value;
        }
        return undefined;
    }
    findIndex(p, thisArg) {
        const boundP = p.bind(thisArg);
        for (const [index, value] of this.entries()) {
            if (!boundP(value, index, this))
                return index;
        }
        return -1;
    }
    findLast(p, thisArg) {
        const boundP = p.bind(thisArg);
        for (let i = this.length - 1; i >= 0; i--) {
            const value = this.#ops.get(i);
            if (value !== undefined && boundP(value, i, this))
                return value;
        }
        return undefined;
    }
    findLastIndex(p, thisArg) {
        const boundP = p.bind(thisArg);
        for (let i = this.length - 1; i >= 0; i--) {
            const item = this.#ops.get(i);
            if (item !== undefined && boundP(item, i, this))
                return i;
        }
        return -1;
    }
    forEach(f, thisArg) {
        Array.prototype.forEach.call(this.#proxy, f, thisArg);
    }
    *keys() {
        for (let i = 0; i < this.length; i++) {
            if (this.#ops.has(i))
                yield i;
        }
    }
    map(f, thisArg) {
        return Array.prototype.map.call(this.#proxy, f, thisArg);
    }
    /** `pop()` works differently from `Array.prototype.pop()`. It returns
     * the last element of the array and leave a hole behind. Since the
     * array have a fixed size, it will not change the value of {@link
     * length}.
     */
    pop() {
        for (let i = this.length - 1; i >= 0; i--) {
            const value = this.#ops.get(i);
            if (value !== undefined) {
                this.#ops.delete(i);
                return value;
            }
        }
        return undefined;
    }
    /** `push()` works differently from `Array.prototype.push()`. It places
     * values in holes following the last element of the container, and
     * returns an array of values that couldn't be placed. Since the array
     * has a fixed size, it will not change the value of {@link length}.
     */
    push(...values) {
        for (let i = this.length; i >= 0; i--) {
            if (this.#ops.has(i)) {
                i++;
                let j = 0;
                for (; j < values.length; j++, i++)
                    this.#ops.set(i, values[j]);
                return values.slice(j + 1);
            }
        }
        return values;
    }
    reduce(f, init) {
        return Array.prototype.reduce.call(this.#proxy, f, init);
    }
    reduceRight(f, init) {
        return Array.prototype.reduceRight.call(this.#proxy, f, init);
    }
    reverse() {
        Array.prototype.reverse.call(this.#proxy);
        return this;
    }
    /** `shift()` works differently from `Array.prototype.shift()`. It
     * returns the first element and leaves a hole beind. Since containers
     * have a fixed size, it will not change the value of {@link length}.
     */
    shift() {
        for (let i = 0; i < this.length; i++) {
            const value = this.#ops.get(i);
            if (value !== undefined) {
                this.#ops.delete(i);
                return value;
            }
        }
        return undefined;
    }
    slice(start, end) {
        return Array.prototype.slice.call(this.#proxy, start, end);
    }
    some(p, thisArg) {
        const boundP = p.bind(thisArg);
        for (const [index, value] of this.entries()) {
            if (boundP(value, index, this))
                return true;
        }
        return false;
    }
    sort(cmp) {
        Array.prototype.sort.call(this.#proxy, cmp);
        return this;
    }
    splice(...args) {
        // @ts-ignore: TypeScript can't prove this is well-typed.
        return Array.prototype.splice.apply(this.#proxy, args);
    }
    /** `unshift()` works differently from `Array.prototype.unshift()`. It
     * places values in holes preceding the first element of the container,
     * and returns an array of values that couldn't be placed. Since
     * containers have a fixed size, it will not change the value of {@link
     * length}.
     */
    unshift(...values) {
        for (let i = 0; i < this.length; i++) {
            if (this.#ops.has(i)) {
                i -= values.length;
                const numSkipped = -i;
                i = Math.max(i, 0);
                for (let j = numSkipped; j < values.length; j++, i++)
                    this[i] = values[j];
                return values.slice(0, numSkipped);
            }
        }
        return values;
    }
    *values() {
        for (let i = 0; i < length; i++) {
            const value = this[i];
            if (value !== undefined)
                yield value;
        }
    }
}
//# sourceMappingURL=array-like.js.map
