import { ItemComponent } from "./component.js";
import * as I from "../inspect.js";
import * as PP from "../pprint.js";
export class ItemDurability extends ItemComponent {
    static typeId = "minecraft:durability";
    #enchantments;
    /// @internal
    constructor(raw, enchantments) {
        super(raw);
        this.#enchantments = enchantments;
    }
    /** Return the current durability of the item. Items will zero
     * durability will likely to break on the next use.
     */
    get current() {
        return this.raw.maxDurability - this.raw.damage;
    }
    /** Update the current durability of the item.
     */
    set current(durability) {
        this.raw.damage = this.raw.maxDurability - durability;
    }
    /** Return the maximum amount of damage that this item can take before
     * breaking.
     */
    get maximum() {
        return this.raw.maxDurability;
    }
    /** Return the chance that this item would be damaged, taking account
     * of its Unbreaking level. The returned value is in the range of [0, 1].
     */
    get damageChance() {
        const level = this.#enchantments.get("unbreaking") ?? 0;
        const chance = this.raw.getDamageChance(level);
        return chance / 100.0;
    }
    /** Apply a damage to the tool, taking account of its Unbreaking
     * level. The damage should usually be 1 for its primary use, and 2 for
     * non-primary uses (like using pickaxes for digging dirt). Return
     * `true` if it exceeded the maximum damage the item can take, i.e. it
     * should break. `false` otherwise.
     */
    damage(amount) {
        if (Math.random() < this.damageChance) {
            if (this.current >= amount) {
                this.current -= amount;
                return false;
            }
            else {
                this.current = 0;
                return true;
            }
        }
        return false;
    }
    /// @internal
    [I.customInspectSymbol](inspect, stylise) {
        const obj = {
            current: this.current,
            maximum: this.maximum,
        };
        try {
            if (this.damageChance < 1)
                obj.damageChance = this.damageChance;
        }
        catch (e) {
            if (I.looksLikeReadonlyError(e))
                // ItemDurabilityComponent.prototype.getDamageChance() isn't
                // callable in read-only mode.
                Object.defineProperty(obj, "damageChance", {
                    get: () => this.damageChance,
                    enumerable: true
                });
            else
                throw e;
        }
        return PP.spaceCat(stylise(PP.text("ItemDurability"), I.TokenType.Class), inspect(obj));
    }
}
//# sourceMappingURL=durability.js.map
