/** Create a proxy object with a thunk to compute its target. As soon as
 * any of their properties are accessed, the thunk is evaluated and the
 * result is memorised.
 */
export function lazy(thunk) {
    let isEvaluated = false;
    let value;
    return new Proxy({}, {
        apply(_targ, _this, args) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return value.call(value, ...args);
        },
        construct(_targ, args) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return new value(...args);
        },
        defineProperty(_targ, key, desc) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            Object.defineProperty(value, key, desc);
            return true;
        },
        deleteProperty(_targ, key) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            delete value[key];
            return true;
        },
        "get"(_targ, key) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            const prop = value[key];
            if (typeof prop === "function")
                // If the property is a function, we need to recover "this"
                // or it won't function as a method.
                return prop.bind(value);
            else
                return prop;
        },
        getOwnPropertyDescriptor(_targ, key) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return Object.getOwnPropertyDescriptor(value, key);
        },
        getPrototypeOf(_targ) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return Object.getPrototypeOf(value);
        },
        has(_targ, key) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return key in value;
        },
        isExtensible(_targ) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return Object.isExtensible(value);
        },
        ownKeys(_targ) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            return Reflect.ownKeys(value);
        },
        preventExtensions(_targ) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            Object.preventExtensions(value);
            return true;
        },
        "set"(_targ, key, v) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            value[key] = v;
            return true;
        },
        setPrototypeOf(_targ, proto) {
            if (!isEvaluated) {
                value = thunk();
                isEvaluated = true;
            }
            Object.setPrototypeOf(value, proto);
            return true;
        }
    });
}
//# sourceMappingURL=lazy.js.map
