import { Queue } from "./collections/queue.js";
import { UniString } from "./unicode.js";
/// Always-backtracking monadic parser. Mostly a port of
/// https://hackage.haskell.org/package/attoparsec but in a totally
/// different implementation. Specifically it doesn't allow partial inputs
/// and isn't implemented in CPS style, otherwise we run out of stack
/// space.
export class Parser {
    run;
    constructor(run) {
        this.run = run;
    }
    then(next) {
        return new Parser((input, pos) => {
            const res = this.run(input, pos);
            if ("value" in res) {
                return next(res.value).run(input, res.pos);
            }
            else {
                return res;
            }
        });
    }
    orElse(alt) {
        return new Parser((input, pos) => {
            const res = this.run(input, pos);
            if ("value" in res) {
                return res;
            }
            else {
                return alt.run(input, pos);
            }
        });
    }
    label(label) {
        return new Parser((input, pos) => {
            const res = this.run(input, pos);
            if ("value" in res) {
                return res;
            }
            else {
                return { ...res, stack: res.stack.cons(label) };
            }
        });
    }
    map(f) {
        return new Parser((input, pos) => {
            const res = this.run(input, pos);
            if ("value" in res) {
                return { ...res, value: f(res.value) };
            }
            else {
                return res;
            }
        });
    }
}
/// Run a parser and return the result. The result may indicate that the
/// parsing has failed.
export function parse(parser, input) {
    const inputU = new UniString(input);
    const res = parser.run(inputU, 0);
    const rest = inputU.substring(res.pos).valueOf();
    if ("value" in res) {
        return { done: res.value, rest };
    }
    else {
        return { failed: res.msg, rest, stack: Array.from(res.stack) };
    }
}
/** Extract a parsed value out of a result. Throws if the result isn't
 * `done`.
 */
export function getResult(res) {
    if ("failed" in res) {
        if (res.stack.length === 0) {
            throw new Error(res.failed);
        }
        else {
            const ctx = [];
            for (const label of res.stack) {
                if (ctx.length > 0) {
                    ctx.push(" > ");
                }
                ctx.push(label);
            }
            throw new Error(`${ctx.join("")}: ${res.failed}`);
        }
    }
    else {
        return res.done;
    }
}
/// A parser that always succeeds without consuming input.
export function pure(value) {
    return new Parser((_input, pos) => {
        return { pos, value };
    });
}
/// A parser that always fails without consuming input.
export function fail(msg) {
    return new Parser((_input, pos) => {
        return { pos, stack: Queue.empty, msg };
    });
}
/** Return all the remaining input without consuming anything. */
const getRemaining = new Parser((input, pos) => {
    return { pos, value: input.substring(pos) };
});
/** Get the current position. */
const getPos = new Parser((_input, pos) => {
    return { pos, value: pos };
});
/** This parser always succeeds. It returns `true` if any input is
 * available, and `false` if the end of all input has been reached.
 */
const hasInput = new Parser((input, pos) => {
    return { pos, value: pos < input.length };
});
/** If at least `n` characters of input are available, return the first `n`
 * characters, otherwise fail. It doesn't consume input.
 */
function ensure(n) {
    return new Parser((input, pos) => {
        if (input.length >= pos + n) {
            return { pos, value: input.substring(pos, pos + n) };
        }
        else {
            return {
                pos,
                stack: Queue.empty,
                msg: `Expected ${n} more characters but reached the end of input`
            };
        }
    });
}
/** Consume `n` characters of input, assuming there indeed is enough input.
 */
function advance(n) {
    return new Parser((_input, pos) => {
        return { pos: pos + n, value: null };
    });
}
/** Match any character but don't consume any input. Succeed with
 * `undefined` if no more input is available.
 */
export const peekChar = new Parser((input, pos) => {
    if (input.length >= pos + 1) {
        return { pos, value: input.at(pos) };
    }
    else {
        return { pos, value: undefined };
    }
});
/// Match any character but don't consume any input. Fail if no more input is
/// available.
export const peekCharOrFail = new Parser((input, pos) => {
    if (input.length >= pos + 1) {
        return { pos, value: input.at(pos) };
    }
    else {
        return { pos, stack: Queue.singleton("peekCharOrFail"), msg: "unexpected end of input" };
    }
});
/// Match any character.
export const anyChar = satisfy(() => true);
/** Succeed for any character which the predicate `p` returns
 * `true`. Return the character that is actually parsed.
 */
export function satisfy(p) {
    return peekCharOrFail.then(c => {
        if (p(c)) {
            return advance(1).then(() => pure(c));
        }
        else {
            return fail("satisfy");
        }
    });
}
/// Parse a single ASCII digit.
export const asciiDigit = satisfy(isAsciiDigit);
/// Return `true` iff the given string contains a single ASCII digit.
export function isAsciiDigit(c) {
    if (c.length === 1) {
        const cp = c.codePointAt(0);
        return cp >= 0x30 && cp <= 0x39;
    }
    else {
        return false;
    }
}
/// Parse a single ASCII alphabet.
export const asciiAlpha = satisfy(isAsciiAlpha);
/// Return `true` iff the given string contains a single ASCII alphabet.
export function isAsciiAlpha(c) {
    if (c.length === 1) {
        const cp = c.codePointAt(0);
        return (cp >= 0x41 && cp <= 0x5A) || (cp >= 0x61 && cp <= 0x7A);
    }
    else {
        return false;
    }
}
/// Parse a single ASCII whitespace, namely SP, HT, LF, VT, FF, CR.
export const asciiSpace = satisfy(isAsciiSpace);
/// Return `true` iff the given string contains a single ASCII whitespace.
export function isAsciiSpace(c) {
    if (c.length === 1) {
        const cp = c.codePointAt(0);
        return (cp === 0x20) || (cp >= 0x09 && cp <= 0x0D);
    }
    else {
        return false;
    }
}
/// Parse a single character matching the given `RegExp`.
export function match(re) {
    return satisfy(c => re.test(c));
}
/// Parse a single character that doesn't match the given `RegExp`.
export function noMatch(re) {
    return satisfy(c => !re.test(c));
}
/// Parse a sequence of bytes that identically match the given string.
export function literal(str) {
    const ustr = new UniString(str);
    return ensure(ustr.length).then(got => {
        if (got.equals(ustr))
            return advance(got.length).then(() => pure(got.valueOf()));
        else
            return fail(`expected literal "${str}"`);
    });
}
function countWhile(p, str) {
    let res = 0;
    for (const c of str) {
        if (p(c))
            res++;
        else
            break;
    }
    return res;
}
/// Skip over ASCII white spaces.
export const skipAsciiSpaces = skipWhile(isAsciiSpace);
/// Skip past input for as long as the predicate returns `true`.
export function skipWhile(p) {
    return getRemaining
        .map(input => countWhile(p, input))
        .then(len => advance(len));
}
/// Consume exactly `n` bytes of input.
export function take(len) {
    const n = Math.max(len, 0);
    return ensure(n).then(s => advance(n).then(() => pure(s.valueOf())));
}
/** A stateful scanner. The predicate consumes and transforms a state
 * argument, and each transformed state is passed to successive invocations
 * of the predicate on each byte of the input until one returns `undefined`
 * or the input ends.
 *
 * This parser does not fail. It will return an empty string if the
 * predicate returns `undefined` on the first byte of input.
 */
export function scan(init, next) {
    function go(state, len) {
        return peekChar.then(c => {
            if (c === undefined) {
                return pure(len);
            }
            else {
                const state1 = next(state, c);
                if (state1 === undefined)
                    return pure(len);
                else
                    return go(state1, len + 1);
            }
        });
    }
    return go(init, 0).then(take);
}
/** Consume input as long as the predicate returns `true`, and return the
 * consumed input.
 *
 * This parser does not fail. It will return an empty string if the
 * predicate returns `false` on the first byte of input.
 */
export function takeWhile(p) {
    return getRemaining.then(input => {
        const len = countWhile(p, input);
        return advance(len).then(() => pure(input.substring(0, len).valueOf()));
    });
}
/** Consume input as long as the predicate returns `true`, and return the
 * consumed input.
 *
 * This parser requires the predicate to succeed on at least one byte of
 * input: it will fail if the predicate never returns `true` or if there is
 * no input left.
 */
export function takeWhile1(p) {
    return takeWhile(p).then(str => {
        if (str.length === 0) {
            return fail(`Expected at least one character satisfying ${p}`);
        }
        else {
            return pure(str);
        }
    });
}
/** Consume input as long as the predicate returns `false`
 * (i.e. until it returns `true`), and return the consumed input.
 */
export function takeTill(p) {
    return takeWhile(c => !p(c));
}
/** Consume all remaining input and return it as a single string.
*/
export const takeRest = getRemaining.then(input => advance(input.length).then(() => pure(input.valueOf())));
/** Match either a single newline character `\n`, or a carriage return
 * followed by a newline character @\"\\r\\n\"@.
 */
export const endOfLine = literal("\n").orElse(literal("\r\n"));
/** Parse and decode an unsigned decimal number.
 */
export const decimal = takeWhile1(isAsciiDigit).then(digits => pure(Number.parseInt(digits)));
/** Parse and decode an unsigned hexadecimal number. The hex digits `a`
 * through `f` may be upper or lower case.
 *
 * This parser does not accept a leading `0x` string.
 */
export const hexadecimal = (() => {
    function isHexDigit(c) {
        const cc = c.codePointAt(0);
        return (cc >= 0x30 && cc <= 0x39) ||
            (cc >= 0x41 && cc <= 0x46) ||
            (cc >= 0x61 && cc <= 0x66);
    }
    return takeWhile1(isHexDigit).then(digits => pure(Number.parseInt(digits, 16)));
})();
/** Parse a number with an optional leading `+` or `-` sign character.
 */
export function signed(p) {
    return choice([
        literal("-").then(() => p.map(n => -1 * n)),
        literal("+").then(() => p),
        p
    ]);
}
/** Parse a floating point number, e.g. `-3.25e-3`. */
export const float = (() => {
    const mantissa = takeWhile(isAsciiDigit).then(integral => peekChar.then(maybeDot => {
        if (maybeDot == ".") {
            return advance(1).then(() => takeWhile(isAsciiDigit).then(frac => pure(`${integral}.${frac}`)));
        }
        else {
            return pure(integral);
        }
    }));
    const maybeExponent = satisfy(c => c === "e" || c === "E").then(() => signed(decimal).then(exp => pure(`E${exp}`)));
    const unsignedFloat = mantissa.then(m => maybeExponent.then(e => pure(Number.parseFloat(m + e))));
    return signed(unsignedFloat);
})();
/** `choice(ps)` tries to apply the parsers in the Iterable `ps` in order,
 * until one of them succeeds. Return the value of the succeeding parser.
 */
export function choice(ps) {
    let ret = fail("no choices were given");
    for (const p of ps) {
        ret = ret.orElse(p);
    }
    return ret.orElse(fail("no choices were taken"));
}
/** Apply the given parser repeatedly, returning every result. */
export function count(n, p) {
    function go(n, acc) {
        if (n > 0) {
            return p.then(v => go(n - 1, acc.snoc(v)));
        }
        else {
            return pure(acc);
        }
    }
    return go(n, Queue.empty).then(vs => pure(Array.from(vs)));
}
/** Try to apply the parser `p`. If the parser fails it returns the value
 * `v`, otherwise the value returned by `p`.
 */
export function option(v, p) {
    return p.orElse(pure(v));
}
/** Apply the parser `p` *zero* or more times. Return an array of the
 * returned values of `p`. The parser throws if `p` succeeds without
 * consuming input, which leads to an infinite loop.
 */
export function many(p) {
    function go(acc, pos0) {
        return p.then(value => {
            return getPos.then(pos1 => {
                if (pos1 === pos0) {
                    throw new Error("Applied to a parser that succeeds without consuming input, " +
                        "which leads to an infinite loop");
                }
                else {
                    return go(acc.snoc(value), pos1);
                }
            });
        }).orElse(pure(Array.from(acc)));
    }
    return getPos.then(pos0 => go(Queue.empty, pos0));
}
/** Apply the parser `p` *one* or more times. Return an array of the
 * returned values of `p`. The parser throws if `p` succeeds without
 * consuming input, which leads to an infinite loop.
 */
export function many1(p) {
    return p.then(value => many(p).then(values => {
        values.unshift(value);
        return pure(values);
    }));
}
/** Apply the parser `p` *zero* or more times until the parser `end`
 * succeeds, and return an array of values returned by `p`. The parser
 * throws if `p` succeeds without consuming input, which leads to an
 * infinite loop.
 */
export function manyTill(p, end) {
    function go(pos0, acc) {
        return end.then(() => pure(Array.from(acc)))
            .orElse(p.then(value => getPos.then(pos1 => {
            if (pos1 === pos0) {
                throw new Error("Applied to a parser that succeeds without consuming input, " +
                    "which leads to an infinite loop");
            }
            else {
                return go(pos1, acc.snoc(value));
            }
        })));
    }
    return getPos.then(pos0 => go(pos0, Queue.empty));
}
/** Apply *zero* or more occurrences of `p`, separated by `sep`. Return an
 * array of the values returned by `p`. The parser throws if both `p` and
 * `sep` succeed without consuming input, which leads to an infinite loop.
 */
export function sepBy(p, sep) {
    return p.then(value => sep.then(() => sepBy1(p, sep))
        .orElse(pure([]))
        .then(values => {
        values.unshift(value);
        return pure(values);
    }))
        .orElse(pure([]));
}
/** Apply *one* or more occurrences of `p`, separated by `sep`. Return an
 * array of the values returned by `p`. The parser throws if both `p` and
 * `sep` succeed without consuming input, which leads to an infinite loop.
 */
export function sepBy1(p, sep) {
    function go(pos0, acc) {
        return p.then(value => sep.then(() => getPos.then(pos1 => {
            if (pos1 === pos0) {
                throw new Error("Applied to parsers that succeed without consuming input, " +
                    "which leads to an infinite loop");
            }
            else {
                // When this call of go() fails, sep too fails and
                // its consumption is canceled.
                return go(pos1, acc.snoc(value));
            }
        }))
            .orElse(pure(Array.from(acc.snoc(value)))));
    }
    return getPos.then(pos0 => go(pos0, Queue.empty));
}
/* This is like {@link many} but discards the result. The parser throws if
 * `p` succeeds without consuming input, which leads to an infinite loop.
 */
export function skipMany(p) {
    function go(pos0) {
        return p.then(() => getPos.then(pos1 => {
            if (pos1 === pos0) {
                throw new Error("Applied to a parser that succeeds without consuming input, " +
                    "which leads to an infinite loop");
            }
            else {
                return go(pos1);
            }
        }))
            .orElse(pure(null));
    }
    return getPos.then(go);
}
/* This is like {@link many1} but discards the result. The parser throws if
 * `p` succeeds without consuming input, which leads to an infinite loop.
 */
export function skipMany1(p) {
    return p.then(() => skipMany(p));
}
/** Match only if all input has been consumed. */
export const endOfInput = hasInput.then(has => {
    if (has) {
        return fail("Expected the end of input but it's not");
    }
    else {
        return pure(null);
    }
});
/** Return an indication of whether the end of input has been reached.
 */
export const atEnd = hasInput.map(b => !b);
//# sourceMappingURL=parser.js.map
