import { Dimension } from "./dimension.js";
import { Entity } from "./entity.js";
import { EntityHealth, EntityLavaMovement, EntityMovement, EntityUnderwaterMovement } from "./entity/attributes.js";
import { EntityBreathable } from "./entity/breathable.js";
import { EntityEquipment } from "./entity/equipment.js";
import { EntityInventory } from "./entity/inventory.js";
import { EntityRideable } from "./entity/rideable.js";
import { map } from "./iterable.js";
import { Location } from "./location.js";
import { PlayerConsole } from "./player/console.js";
import { Wrapper } from "./wrapper.js";
import { GameMode } from "@minecraft/server";
import * as Prefs from "./preferences.js";
import * as I from "./inspect.js";
import * as PP from "./pprint.js";
export { GameMode };
export { ScreenDisplay, PlayerLeaveAfterEvent } from "@minecraft/server";
export class Player extends Entity {
    #session;
    /** Package private: user code should not use this. */
    constructor(rawPlayer) {
        super(rawPlayer);
    }
    /** Package private: user code should not use this. */
    get rawPlayer() {
        // NOTE: We cannot simply override "get raw()", possibly due to
        // https://github.com/microsoft/TypeScript/issues/41347
        return super.raw;
    }
    /** A Console API that sends messages to the chat screen of this
     * player. The message will also be sent to the content log. This is
     * mainly for error reporting and debugging. Use {@link sendMessage}
     * for regular messages. */
    get console() {
        return new PlayerConsole(this);
    }
    get gameMode() {
        // It is surprising that this is the only way to obtain the
        // per-player game mode. Hope the future API will allow us to do it
        // efficiently.
        if (this.matches({ gameMode: GameMode.adventure }))
            return GameMode.adventure;
        if (this.matches({ gameMode: GameMode.creative }))
            return GameMode.creative;
        if (this.matches({ gameMode: GameMode.spectator }))
            return GameMode.spectator;
        if (this.matches({ gameMode: GameMode.survival }))
            return GameMode.survival;
        throw new Error(`Cannot detect the game mode for player ${this.name}`);
    }
    get isEmoting() {
        return this.rawPlayer.isEmoting;
    }
    get isFlying() {
        return this.rawPlayer.isFlying;
    }
    get isGliding() {
        return this.rawPlayer.isGliding;
    }
    get isJumping() {
        return this.rawPlayer.isJumping;
    }
    get isOp() {
        return this.rawPlayer.isOp();
    }
    set isOp(b) {
        this.rawPlayer.setOp(b);
    }
    get level() {
        return this.rawPlayer.level;
    }
    get name() {
        return this.rawPlayer.name;
    }
    get onScreenDisplay() {
        return this.rawPlayer.onScreenDisplay;
    }
    get selectedSlotIndex() {
        return this.rawPlayer.selectedSlotIndex;
    }
    set selectedSlotIndex(slot) {
        this.rawPlayer.selectedSlotIndex = slot;
    }
    get spawnPoint() {
        const raw = this.rawPlayer.getSpawnPoint();
        if (raw)
            return {
                dimension: new Dimension(raw.dimension),
                location: new Location(raw.x, raw.y, raw.z),
            };
        else
            return undefined;
    }
    set spawnPoint(dl) {
        if (dl)
            this.rawPlayer.setSpawnPoint({
                dimension: dl.dimension.raw,
                x: dl.location.x,
                y: dl.location.y,
                z: dl.location.z,
            });
        else
            this.rawPlayer.setSpawnPoint();
    }
    get totalXp() {
        return this.rawPlayer.getTotalXp();
    }
    get totalXpNeededForNextLevel() {
        return this.rawPlayer.totalXpNeededForNextLevel;
    }
    get xpEarnedAtCurrentLevel() {
        return this.rawPlayer.xpEarnedAtCurrentLevel;
    }
    /** Add/remove experience to/from the player and return their current
     * experience. The amount can be negative.
     */
    addExperience(amount) {
        return this.rawPlayer.addExperience(amount);
    }
    /** Add/remove level to/from the player and return their current
     * level. The amount can be negative.
     */
    addLevels(amount) {
        return this.rawPlayer.addLevels(amount);
    }
    /** Obtain the per-player preferences object for this player. */
    getPreferences(ty) {
        return Prefs.decodeOrCreate(ty, this.getDynamicProperty(Prefs.dynamicPropertyId("player"), "string?"));
    }
    /** Update the per-player preferences object for this player. */
    setPreferences(ty, prefs) {
        this.setDynamicProperty(Prefs.dynamicPropertyId("player"), Prefs.encode(ty, prefs));
    }
    /** Obtain a player session object. To use this method, you first need
     * to register your session class using {@link
     * World.prototype.usePlayerSessions}.
     */
    getSession() {
        if (this.#session === undefined)
            this.#session = SessionManager.get(this.id);
        return this.#session;
    }
    playSound(soundId, soundOptions) {
        this.rawPlayer.playSound(soundId, soundOptions);
    }
    resetLevel() {
        this.rawPlayer.resetLevel();
    }
    sendMessage(msg) {
        this.rawPlayer.sendMessage(msg);
    }
    // --------- Components ---------
    // These components should always exist.
    get breathable() {
        return super.breathable;
    }
    get equipment() {
        return super.equipment;
    }
    get health() {
        return super.health;
    }
    get inventory() {
        return super.inventory;
    }
    get lavaMovement() {
        return super.lavaMovement;
    }
    get movement() {
        return super.movement;
    }
    get rideable() {
        return super.rideable;
    }
    get underwaterMovement() {
        return super.underwaterMovement;
    }
    /// @internal
    [I.customInspectSymbol](inspect, stylise) {
        const obj = {
            name: this.name,
            dimension: this.dimension,
            location: this.location,
            experience: {
                level: this.level,
                totalXp: this.totalXp,
                totalXpNeededForNextLevel: this.totalXpNeededForNextLevel,
                xpEarnedAtCurrentLevel: this.xpEarnedAtCurrentLevel,
            },
            selectedSlotIndex: this.selectedSlotIndex,
        };
        if (this.isSneaking)
            obj.isSneaking = true;
        if (this.isEmoting)
            obj.isEmoting = true;
        if (this.isFlying)
            obj.isFlying = true;
        if (this.isGliding)
            obj.isGliding = true;
        if (this.isJumping)
            obj.isJumping = true;
        if (this.isOp)
            obj.isOp = true;
        if (this.spawnPoint)
            obj.spawnPoint = this.spawnPoint;
        if (this.tags.size > 0)
            obj.tags = this.tags;
        // These are flags but are in fact components.
        if (this.canClimb)
            obj.canClimb = true;
        if (this.isHiddenWhenInvisible)
            obj.isHiddenWhenInvisible = true;
        const comps = new Set([
            this.breathable,
            this.health,
            this.inventory,
            this.lavaMovement,
            this.movement,
            this.rideable,
            this.underwaterMovement,
        ]);
        try {
            if (this.equipment.size > 0)
                comps.add(this.equipment);
        }
        catch (e) {
            if (I.looksLikeReadonlyError(e))
                // EntityEquippableComponent.prototype.getEquipment() isn't
                // callable in read-only mode.
                comps.add(this.equipment);
            else
                throw e;
        }
        for (const comp of this.raw.getComponents()) {
            switch (comp.typeId) {
                case EntityBreathable.typeId:
                case "minecraft:can_climb":
                case EntityHealth.typeId:
                case EntityEquipment.typeId:
                case EntityInventory.typeId:
                case "minecraft:is_hidden_when_invisible":
                case EntityLavaMovement.typeId:
                case EntityMovement.typeId:
                case EntityRideable.typeId:
                case EntityUnderwaterMovement.typeId:
                    // These are already inspected in our own way.
                    break;
                default:
                    comps.add(comp);
            }
        }
        if (comps.size > 0)
            obj.components = comps;
        return PP.spaceCat(stylise(PP.text("Player"), I.TokenType.Class), inspect(obj));
    }
}
/// @internal
export class SessionManager {
    static #sessions = new Map(); // playerId => IPlayerSession
    static #ctor;
    static get "class"() {
        return this.#ctor;
    }
    static set "class"(ctor) {
        if (this.#ctor)
            throw new Error("A session class cannot be changed once it's set");
        else
            this.#ctor = ctor;
    }
    static create(player) {
        if (!this.#ctor)
            throw new Error("No session classes have been set");
        if (this.#sessions.has(player.id))
            throw new Error(`Duplicate session for player ${player.id} (${player.name})`);
        const session = new (this.#ctor)(player);
        this.#sessions.set(player.id, session);
        return session;
    }
    static destroy(playerId) {
        const session = this.#sessions.get(playerId);
        if (session) {
            this.#sessions.delete(playerId);
            session.destroy();
        }
        else {
            throw new Error(`Session not found for player ${playerId}`);
        }
    }
    static "get"(playerId) {
        const session = this.#sessions.get(playerId);
        if (session) {
            return session;
        }
        else {
            if (this.#ctor)
                throw new Error(`Session not found for player ${playerId}`);
            else
                throw new Error("The session manager hasn't been configured");
        }
    }
}
export class ChatSendBeforeEvent extends Wrapper {
    cancel() {
        this.raw.cancel = true;
    }
    get message() {
        return this.raw.message;
    }
    get sender() {
        return new Player(this.raw.sender);
    }
    get targets() {
        if (this.raw.targets)
            return map(this.raw.targets, raw => new Player(raw));
        else
            return undefined;
    }
}
//# sourceMappingURL=player.js.map
