import { beside, column, empty, text, space, line, linebreak, nest, nesting } from "./primitives.js";
import { group } from "./flatten.js";
/** The document `sep(ds)` concatenates all documents `ds` either
 * horizontally with {@link spaceCat} if it fits the page, or vertically
 * with {@link lineCat}.
 */
export function sep(ds) {
    return group(vsep(ds));
}
/** The document `fillSep(ds)` concatenates documents `ds` horizontally
 * with {@link softlineCat} as long as its fits the page, then inserts a
 * {@link line} and continues doing that for all documents in `ds`.
 */
export function fillSep(ds) {
    return fold(softlineCat, ds);
}
/** The document `hsep(ds)` concatenates all documents `ds` horizontally
 * with {@link spaceCat}.
 */
export function hsep(ds) {
    return fold(spaceCat, ds);
}
/** The document `vsep(ds)` concatenates all documents `ds` vertically with
 * {@link lineCat}. If a {@link group} undoes the line breaks inserted by
 * `vsep`, all documents are separated with a space.
 *
 * ``javascript
 * const someText = ["text", "to", "lay", "out"].map(text);
 * const test     = spaceCat(text("some"), vsep(someText));
 * ``
 *
 * This is layed out as:
 *
 * ``
 * some text
 * to
 * lay
 * out
 * ``
 *
 * The {@link align} combinator can be used to align the documents under
 * their first element
 *
 * ``javascript
 * const test = spaceCat(text("some"), align(vsep(someText)));
 * ``
 *
 * Which is printed as:
 *
 * ``
 * some text
 *      to
 *      lay
 *      out
 * ``
 */
export function vsep(ds) {
    return fold(lineCat, ds);
}
/** The document `cat(ds)` concatenates all documents `ds` either
 * horizontally with {@link beside} if it fits the page, or vertically with
 * {@link linebreakCat}.
 */
export function cat(ds) {
    return group(vcat(ds));
}
/** The document `fillCat(ds)` concatenates documents `ds` horizontally
 * with {@link beside} as long as its fits the page, then inserts a {@link
 * linebreak} and continues doing that for all documents in `ds`.
 */
export function fillCat(ds) {
    return fold(softbreakCat, ds);
}
/** The document `hcat(ds)` concatenates all documents `ds` horizontally
 * with {@link beside}.
 */
export function hcat(ds) {
    return fold(beside, ds);
}
/** The document `vcat(ds)` concatenates all documents `ds` vertically with
 * {@link linebreakCat}. If a {@link group} undoes the line breaks inserted
 * by `vcat`, all documents are directly concatenated.
 */
export function vcat(ds) {
    return fold(linebreakCat, ds);
}
/** Fold an array of documents with a given binary operator. If the array
 * is empty, the resulting document will also be {@link empty}.
 */
export function fold(f, ds) {
    return ds.length == 0
        ? empty
        : ds.reduce(f);
}
/** The document `softline` behaves like {@link space} if the resulting
 * output fits the page, otherwise it behaves like {@link line}.
 */
export const softline = group(line);
/** The document `softbreak` behaves like {@link empty} if the resulting
 * output fits the page, otherwise it behaves like {@link line}.
 */
export const softbreak = group(linebreak);
/** The document `lparen` contains a left parenthesis `(`.
 */
export const lparen = text("(");
/** The document `lparen` contains a right parenthesis `)`.
 */
export const rparen = text(")");
/** The document `langle` contains a left angle bracket `<`.
 */
export const langle = text("<");
/** The document `rangle` contains a right angle bracket `>`.
 */
export const rangle = text(">");
/** The document `lbrace` contains a left brace `{`.
 */
export const lbrace = text("{");
/** The document `rbrace` contains a right brace `}`.
 */
export const rbrace = text("}");
/** The document `lbracket` contains a left square bracket `[`.
 */
export const lbracket = text("[");
/** The document `rbracket` contains a right square bracket `]`.
 */
export const rbracket = text("]");
/** The document `squote` contains a single quote `'`.
 */
export const squote = text("'");
/** The document `dquote` contains a double quote `"`.
 */
export const dquote = text('"');
/** The document `semi` contains a semicolon `;`.
 */
export const semi = text(";");
/** The document `colon` contains a colon `:`.
 */
export const colon = text(":");
/** The document `colon` contains a comma `,`.
 */
export const comma = text(",");
/** The document `colon` contains a single dot `.`.
 */
export const dot = text(".");
/** The document `backslash` contains a back slash `\`.
 */
export const backslash = text("\\");
/** The document `equals` contains an equal sign `=`.
 */
export const equals = text("=");
/** The document `spaceCat(x, y)` concatenates document `x` and `y` with a
 * {@link space} in between.
 */
export function spaceCat(x, y) {
    return beside(beside(x, space), y);
}
/** The document `softlineCat(x, y)` concatenates document `x` and `y` with
 * a {@link softline} in between. This effectively puts `x` and `y` either
 * next to each other (with a {@link space} in between) or underneath each
 * other.
 */
export function softlineCat(x, y) {
    return beside(beside(x, softline), y);
}
/** The document `softbreakCat(x, y)` concatenates document `x` and `y`
 * with a {@link softbreak} in between. This effectively puts `x` and `y`
 * either right next to each other or underneath each other.
 */
export function softbreakCat(x, y) {
    return beside(beside(x, softbreak), y);
}
/** The document `lineCat(x, y)` concatenates document `x` and `y` with a
 * {@link line} in between.
 */
export function lineCat(x, y) {
    return beside(beside(x, line), y);
}
/** The document `linebreakCat(x, y)` concatenates document `x` and `y`
 * with a {@link linebreak} in between.
 */
export function linebreakCat(x, y) {
    return beside(beside(x, linebreak), y);
}
/** Document `squotes(d)` encloses document `d` with single quotes `'`.
 */
export function squotes(d) {
    return enclose(squote, squote, d);
}
/** Document `dquotes(d)` encloses document `d` with double quotes `"`.
 */
export function dquotes(d) {
    return enclose(dquote, dquote, d);
}
/** Document `braces(d)` encloses document `d` in braces `{` and `}`.
 */
export function braces(d) {
    return enclose(lbrace, rbrace, d);
}
/** Document `parens(d)` encloses document `d` in parentheses `(` and `)`.
 */
export function parens(d) {
    return enclose(lparen, rparen, d);
}
/** Document `angles(d)` encloses document `d` in angle brackets `<` and
 * `>`.
 */
export function angles(d) {
    return enclose(langle, rangle, d);
}
/** Document `brackets(d)` encloses document `d` in square brackets `[` and
 * `]`.
 */
export function brackets(d) {
    return enclose(lbracket, rbracket, d);
}
/** The document `enclose(l, r, d)` encloses document `d` between documents
 * `l` and `r` using {@link beside}.
 */
export function enclose(l, r, d) {
    return beside(beside(l, d), r);
}
/** The document `string(s)` concatenates all characters in `s` using
 * {@link line} for newline characters and {@link text} for all other
 * characters. It is used instead of {@link text} whenever the text
 * contains newline characters.
 */
export function string(s) {
    const ds = [];
    let begin = 0;
    while (true) {
        const pos = s.indexOf("\n", begin);
        if (pos == -1) {
            ds.push(text(s.slice(begin)));
            break;
        }
        else {
            ds.push(text(s.slice(begin, pos)), line);
            begin = pos + 1;
        }
    }
    return hcat(ds);
}
/** The document `bool(b)` shows the literal boolean `b` using {@link
 * text}.
 */
export function bool(b) {
    return text(String(b));
}
/** The document `number(n)` shows the literal number `n` using {@link
 * text}.
 */
export function number(n) {
    return text(String(n));
}
export function spaces(n) {
    return n <= 0 ? "" : " ".repeat(n);
}
/** The document `fill(i, d)` renders document `d`. It then appends {@link
 * space}s until the width is equal to `i`. If the width of `d` is already
 * larger, nothing is appended. This combinator is quite useful in practice
 * to output a list of bindings. The following example demonstrates this.
 *
 * ``javascript
 * const types = [ ["empty", "Doc"],
 *                 ["nest", "Int -> Doc -> Doc"],
 *                 ["linebreak", "Doc"] ];
 * function ptype([name, tp]: [string, string]): Doc {
 *     return spaceCat(spaceCat(fill(6, text(name)), text("::")), text(tp));
 * }
 * const test = spaceCat(text("let"), align(vcat(types.map(ptype))));
 * ``
 *
 * Which is layed out as:
 *
 * ``
 * let empty  :: Doc
 *     nest   :: Int -> Doc -> Doc
 *     linebreak :: Doc
 * ``
 */
export function fill(i, d) {
    return width(d, w => {
        if (w >= i) {
            return empty;
        }
        else {
            return text(spaces(i - w));
        }
    });
}
/** The document `fillBreak(i, d)` first renders document `d`. It then
 * appends {@link space}s until the width is equal to `i`. If the width of
 * `d` is already larger than `i`, the nesting level is increased by `i`
 * and a {@link line} is appended. When we redefine `ptype` in the previous
 * example to use `fillBreak`, we get a useful variation of the previous
 * output:
 *
 * ``javascript
 * function ptype([name, tp]: [string, string]) {
 *     return spaceCat(spaceCat(fillBreak(6, text(name)), text("::")), text(tp));
 * }
 * ``
 *
 * The output will now be:
 *
 * ``
 * let empty  :: Doc
 *     nest   :: Int -> Doc -> Doc
 *     linebreak
 *            :: Doc
 * ``
 */
export function fillBreak(i, d) {
    return width(d, w => {
        if (w > i) {
            return nest(i, linebreak);
        }
        else {
            return text(spaces(i - w));
        }
    });
}
function width(d, f) {
    return column(k1 => beside(d, column(k2 => f(k2 - k1))));
}
/** The document `indent(i, d)` indents document `d` with `i` spaces.
 *
 * ``javascript
 * const test = indent(
 *                4,
 *                fillSep("indent combinator indents these words !"
 *                          .split(" ")
 *                          .map(text)));
 * ``
 *
 * Which lays out with a page width of 20 as:
 *
 * ``
 *     the indent
 *     combinator
 *     indents these
 *     words !
 * ``
 */
export function indent(i, d) {
    return hang(i, beside(text(spaces(i)), d));
}
/** The hang combinator implements hanging indentation. The document
 * `hang(i, d)` renders document `d` with a nesting level set to the
 * current column plus `i`. The following example uses hanging indentation
 * for some text:
 *
 * ``javascript
 * const test = hang(
 *                4,
 *                fillSep("the hang combinator indents these words !"
 *                          .split(" ")
 *                          .map(text)));
 * ``
 *
 * Which lays out on a page with a width of 20 characters as:
 *
 * ``
 * the hang combinator
 *     indents these
 *     words !
 * ``
 */
export function hang(i, d) {
    return align(nest(i, d));
}
/** The document `align(d)` renders document `d` with the nesting level set
 * to the current column. It is used for example to implement {@link hang}.
 *
 * As an example, we will put a document right above another one,
 * regardless of the current nesting level:
 *
 * ``javascript
 * function paste(x: Doc, y: Doc): Doc {
 *     return align(lineCat(x, y));
 * }
 * const test = spaceCat(text("hi"), paste(text("nice"), text("world"));
 * ``
 *
 * which will be layed out as:
 *
 * ``
 * hi nice
 *    world
 * ``
 */
export function align(d) {
    return column(k => nesting(i => nest(k - i, d)));
}
//# sourceMappingURL=combinators.js.map
