import { Queue } from "../collections/queue.js";
import { Notify } from "./notify.js";
export class ChannelBase {
    queue;
    sent;
    constructor() {
        this.queue = Queue.empty;
        this.sent = new Notify();
    }
    async receive() {
        while (true) {
            if (this.queue.isEmpty) {
                await this.sent.notified();
                continue;
            }
            else {
                const [head, rest] = this.queue.uncons();
                this.queue = rest;
                return head;
            }
        }
    }
    /** If the queue is empty, block until at least one message is
     * sent. Otherwise return all the messages in the queue.
     */
    async receiveAll() {
        while (true) {
            if (this.queue.isEmpty) {
                await this.sent.notified();
                continue;
            }
            else {
                const messages = this.queue;
                this.queue = Queue.empty;
                return messages;
            }
        }
    }
}
/** A bounded multi-producer, single-consumer queue for sending values
 * between threads. The bounded variant has a limit on the number of
 * messages that the channel can store, and if this limit is reached,
 * trying to send another message will wait until a message is received
 * from the channel.
 */
export class Channel extends ChannelBase {
    limit;
    #received;
    constructor(limit) {
        super();
        this.limit = limit;
        this.#received = new Notify();
    }
    async receive() {
        const message = await super.receive();
        this.#received.notifyAll();
        return message;
    }
    async receiveAll() {
        const messages = await super.receiveAll();
        this.#received.notifyAll();
        return messages;
    }
    async send(value) {
        while (true) {
            if (this.queue.length < this.limit) {
                this.queue = this.queue.snoc(value);
                this.sent.notifyAll();
                return;
            }
            else {
                await this.#received.notified();
                continue;
            }
        }
    }
}
/** An unbounded multi-producer, single-consumer queue for sending values
 * between threads. An unbounded channel has an infinite capacity, so the
 * send method will always complete immediately. This makes the
 * `UnboundedChannel` usable from both synchronous and asynchronous code.
 */
export class UnboundedChannel extends ChannelBase {
    send(value) {
        this.queue = this.queue.snoc(value);
        this.sent.notifyAll();
    }
}
//# sourceMappingURL=mpsc.js.map
