var _a;
import { delayTicks } from "./delay.js";
import { Timer } from "./timer.js";
import { system, TicksPerSecond } from "@minecraft/server";
export function spawn(...args) {
    let name;
    let gen;
    switch (args.length) {
        case 1:
            name = args[0].name === "" ? "<anonymous>" : args[0].name;
            gen = args[0];
            break;
        case 2:
            name = args[0];
            gen = args[1];
    }
    class T extends Thread {
        run = gen;
    }
    return new T(name).start();
}
export class Thread {
    #task;
    id;
    name;
    #result;
    #isCancelRequested;
    #cancel;
    // In the past Bedrock would postpone running scheduled promises when
    // they're going to exceed the time budget (50 ms per tick). But things
    // appear to have changed in 1.20.60 when they introduced
    // `system.runJob()`. Now the game appears to run all the scheduled
    // promises before entering the next game tick, and can trigger the
    // watchdog when there are too many scheduled promises even if
    // individual promises are small. This forces us to measure the time
    // and insert delays accordingly.
    static #lastTick = 0;
    static #timeBudgetInMsPerTick = 1000 / TicksPerSecond;
    static #timer = new Timer();
    /** The ID of the next thread to be created. */
    static #nextThreadID = 0;
    static #maybeResetTimer() {
        const currentTick = system.currentTick;
        if (this.#lastTick != currentTick) {
            this.#lastTick = currentTick;
            this.#timer.reset();
        }
    }
    static #hasExceededTimeBudget() {
        this.#maybeResetTimer();
        return this.#timer.elapsedMs >= this.#timeBudgetInMsPerTick;
    }
    constructor(name) {
        this.id = _a.#nextThreadID++;
        this.name = name != null ? name : new.target.name;
        this.#isCancelRequested = false;
        // This will be clobbered when the thread starts running.
        this.#cancel = () => { };
    }
    /** Start the thread. It doesn't start on its own just by constructing
     * an instance, because that means `run()` would be invoked even before
     * constructors of subclasses complete.
     */
    start() {
        // One of the two mechanisms to cancel a thread. The promise is
        // passed to the async generator function and will never be
        // resolved. When a cancellation is requested, the promise will be
        // rejected.
        const cancelled = new Promise((_resolve, reject) => {
            this.#cancel = reject;
        });
        // Attach a no-op handler to the cancellation promise, or the
        // runtime will think its rejection is accidentally unhandled.
        cancelled.catch(() => { });
        // This method might be called in the context of before events,
        // which means we might be in a read-only mode. The task should
        // await (or yield in our case since we ignore any yielded values)
        // at least once so that it can mutate the world state.
        const self = this;
        this.#task = (async function* () {
            yield;
            yield* self.run(cancelled);
        })();
        // Since this.#task is an async generator and we haven't called its
        // .next() even once, the generator isn't running yet even
        // asynchronously. Schedule its execution now.
        //
        // And when the promise is fulfilled, we should continue the
        // execution of the task until it finishes or raises an error.
        _a.#maybeResetTimer();
        this.#result = this.#task.next()
            .then(res => this.#onSuspended(res), e => this.#onError(e));
        return this;
    }
    #onSuspended(res) {
        if (res.done) {
            return Promise.resolve(res);
        }
        else {
            // Throwing an exception into the generator would work most of
            // the time, unless the generator is awaiting a promise. In
            // that case, what happens is that the generator yields an
            // unfulfilled promise and the execution stops at "cont.then()"
            // below, which won't be interrupted by this.#task.throw(). So
            // when a generator does "yield await", it has to
            // Promise.race() with the cancellation promise in order to
            // respond to requests in a timely manner.
            let cont;
            if (this.#isCancelRequested) {
                cont = this.#task.throw(new ThreadCancellationRequested());
            }
            else if (_a.#hasExceededTimeBudget()) {
                cont = delayTicks(1).then(() => this.#task.next());
            }
            else {
                cont = this.#task.next();
            }
            return cont.then(res => this.#onSuspended(res), e => this.#onError(e));
        }
    }
    #onError(e) {
        if (e instanceof ThreadCancellationRequested) {
            // The thread didn't catch the cancellation request, which is
            // perfectly fine.
        }
        else {
            console.error("Thread #%d (%s) aborted: %o", this.id, this.name, e);
        }
        // NOTE: Returning a rejected promise here will make join() reject,
        // which is not desirable for us.
        return Promise.resolve({ value: undefined, done: true });
    }
    /** Return a promise which will be resolved when the thread terminates
     * either by exiting normally or throwing an error. Getting a
     * cancellation request and not catching it also counts as throwing an
     * error. If the thread has already terminated, the resulting promise
     * will be an already resolved one.
     *
     * Note that unlike the POSIX threading API, it is perfectly fine to
     * join a thread twice or more. Subsequent joins will just return
     * resolved promises.
     */
    async join() {
        /* We really want to prevent deadlocks by attempting to join a
         * thread from within itself. But the exact moment when the
         * interpreter evaluates the generator isn't under our control so
         * we can't detect deadlocks.
         *
         * Maybe we can take the same abysmal approach as zone.js from
         * Angular but we of course don't want to. Maybe one day we can do
         * better when https://www.proposals.es/proposals/Async%20Context
         * is implemented.
         */
        await this.#result;
    }
    /** Request a cancellation of a thread. The thread is expected to
     * terminate itself shortly, but there's no guarantee of that. This
     * operation is asynchronous, that is, `cancel()` may return before the
     * thread actually terminates. If you want to wait for its termination,
     * call `join()` after this.
     */
    cancel() {
        this.#isCancelRequested = true;
        this.#cancel(); // Reject the cancellation promise.
        return this;
    }
}
_a = Thread;
export class ThreadCancellationRequested extends Error {
}
//# sourceMappingURL=thread.js.map
