import { delay } from "./delay.js";
import { Wrapper } from "./wrapper.js";
import { FormCancelationReason } from "@minecraft/server-ui";
import * as UI from "@minecraft/server-ui";
export { FormCancelationReason };
export class FormResponse extends Wrapper {
    cancelationReason;
    canceled;
    /// @internal
    constructor(raw) {
        super(raw);
        if (raw.cancelationReason)
            this.cancelationReason = raw.cancelationReason;
        this.canceled = raw.canceled;
    }
}
export class ActionFormResponse extends FormResponse {
    selection;
    /// @internal
    constructor(raw, keys) {
        super(raw);
        if (raw.selection !== undefined)
            this.selection = keys[raw.selection];
    }
}
export class MessageFormResponse extends FormResponse {
    selection;
    /// @internal
    constructor(raw, keys) {
        super(raw);
        if (raw.selection !== undefined)
            this.selection = keys[raw.selection];
    }
}
export class ModalFormResponse extends FormResponse {
    formValues;
    /// @internal
    constructor(raw, interpret) {
        super(raw);
        if (raw.formValues) {
            this.formValues = interpret(raw.formValues);
        }
    }
}
export class ModalFormValues extends Map {
    getBoolean(key) {
        const val = this.get(key);
        switch (typeof val) {
            case "boolean":
                return val;
            case undefined:
                throw new Error(`No form values exist for key ${String(key)}`);
            default:
                throw new Error(`The form value for key ${String(key)} is not a boolean: ${String(val)}`);
        }
    }
    getNumber(key) {
        const val = this.get(key);
        switch (typeof val) {
            case "number":
                return val;
            case undefined:
                throw new Error(`No form values exist for key ${String(key)}`);
            default:
                throw new Error(`The form value for key ${String(key)} is not a number: ${String(val)}`);
        }
    }
    getString(key) {
        const val = this.get(key);
        switch (typeof val) {
            case "string":
                return val;
            case undefined:
                throw new Error(`No form values exist for key ${String(key)}`);
            default:
                throw new Error(`The form value for key ${String(key)} is not a string: ${String(val)}`);
        }
    }
}
function unifyToNumber(val) {
    return typeof val === "number"
        ? val
        : (val ? Infinity : 0);
}
async function genericShow(thunk, opts) {
    let busyRetries = 0;
    let cancelRetries = 0;
    while (true) {
        const res = await thunk();
        switch (res.cancelationReason) {
            case undefined: // Submitted
                return res;
            case UI.FormCancelationReason.UserBusy:
                if (busyRetries < unifyToNumber(opts?.retryWhenBusy ?? false)) {
                    busyRetries++;
                    await delay(opts?.retryDelayWhenBusy ?? 0.2);
                    continue;
                }
                return res;
            case UI.FormCancelationReason.UserClosed:
                if (cancelRetries < unifyToNumber(opts?.retryOnCancel ?? false)) {
                    cancelRetries++;
                    continue;
                }
                return res;
        }
    }
}
export class ActionFormData extends Wrapper {
    #buttonKeys;
    constructor() {
        super(new UI.ActionFormData());
        this.#buttonKeys = [];
    }
    /** Sets the body text for the modal form. */
    body(bodyText) {
        this.raw.body(bodyText);
        return this;
    }
    /** Add a button to this form with an icon from a resource pack. The
     * value `key` will show up in {@link
     * ActionFormResponse.prototype.selection} if the form is closed by
     * clicking this button.
     */
    button(key, text, iconPath) {
        this.#buttonKeys.push(key);
        this.raw.button(text, iconPath);
        return this;
    }
    /** Add a section divider to the form. */
    divider() {
        this.raw.divider();
        return this;
    }
    /** Add a header to the form. */
    header(text) {
        this.raw.header(text);
        return this;
    }
    /** Add a label to the form. */
    label(text) {
        this.raw.label(text);
        return this;
    }
    /** Create and show this modal popup form. Return asynchronously when
     * the player confirms or cancels the dialog.
     */
    async show(player, opts) {
        const raw = await genericShow(() => this.raw.show(player.rawPlayer), opts);
        return new ActionFormResponse(raw, this.#buttonKeys);
    }
    /** Set the title for the modal dialog. */
    title(titleText) {
        this.raw.title(titleText);
        return this;
    }
}
export class MessageFormData extends Wrapper {
    #buttonKeys;
    constructor() {
        super(new UI.MessageFormData());
        this.#buttonKeys = [undefined, undefined];
    }
    /** Set the body text for the modal form. */
    body(bodyText) {
        this.raw.body(bodyText);
        return this;
    }
    /** Set the text for the first button of the dialog. The value `key`
     * will show up in {@link MessageFormResponse.prototype.selection} if
     * the form is closed by clicking this button.
     */
    button1(key, text) {
        this.#buttonKeys[0] = key;
        this.raw.button1(text);
        return this;
    }
    /** Set the text for the second button of the dialog. The value `key`
     * will show up in {@link MessageFormResponse.prototype.selection} if
     * the form is closed by clicking this button.
     */
    button2(key, text) {
        this.#buttonKeys[1] = key;
        this.raw.button2(text);
        return this;
    }
    /** Create and show this modal popup form. Return asynchronously when
     * the player confirms or cancels the dialog.
     */
    async show(player, opts) {
        const raw = await genericShow(() => this.raw.show(player.rawPlayer), opts);
        return new MessageFormResponse(raw, this.#buttonKeys);
    }
    /** Set the title for the modal dialog. */
    title(titleText) {
        this.raw.title(titleText);
        return this;
    }
}
export class ModalFormData extends Wrapper {
    #items;
    constructor() {
        super(new UI.ModalFormData());
        this.#items = [];
    }
    /** Add a section divider to the form. */
    divider() {
        this.raw.divider();
        return this;
    }
    /** Add a dropdown with choices to the form. `options` is an iterable
     * object (such as `Array`) of tuples `[optionKey, optionLabel]`. The
     * value `key` will show up as a key of {@link ModalFormResponse.prototype.formValues} with a
     * corresponding option key.
     */
    dropdown(key, label, options, defaultValueIndex) {
        const optKeys = [];
        const optLabels = [];
        for (const [optKey, optLabel] of options) {
            optKeys.push(optKey);
            optLabels.push(optLabel);
        }
        this.#items.push(formValue => {
            if (typeof formValue !== "number")
                throw new Error(`Internal error: wrong type of form value ${String(formValue)} for key ${String(key)}`);
            return [key, optKeys[formValue]];
        });
        this.raw.dropdown(label, optLabels, defaultValueIndex);
        return this;
    }
    /** Add a header to the form. */
    header(text) {
        this.raw.header(text);
        return this;
    }
    /** Add a label to the form. */
    label(text) {
        this.raw.label(text);
        return this;
    }
    /** Create and show this modal popup form. Return asynchronously
     * when the player confirms or cancels the dialog.
     */
    async show(player, opts) {
        const raw = await genericShow(() => this.raw.show(player.rawPlayer), opts);
        return new ModalFormResponse(raw, formValues => {
            const map = new ModalFormValues();
            for (let i = 0; i < this.#items.length; i++) {
                const [key, val] = this.#items[i](formValues[i]);
                map.set(key, val);
            }
            return map;
        });
    }
    /** Add a numeric slider to the form. The value `key` will show up as a
     * key of {@link ModalFormResponse.prototype.formValues} with a
     * corresponding numeric value.
     */
    slider(key, label, minimumValue, maximumValue, valueStep, defaultValue) {
        this.#items.push(formValue => [key, formValue]);
        this.raw.slider(label, minimumValue, maximumValue, valueStep, defaultValue);
        return this;
    }
    /** Set the label for the submit button. */
    submitButton(label) {
        this.raw.submitButton(label);
        return this;
    }
    /** Add a textbox to the form. The value `key` will show up as a key of
     * {@link ModalFormResponse.prototype.formValues} with a corresponding
     * string value.
     */
    textField(key, label, placeholderText, defaultValue) {
        this.#items.push(formValue => [key, formValue]);
        this.raw.textField(label, placeholderText, defaultValue);
        return this;
    }
    /** Set the title for the modal dialog. */
    title(titleText) {
        this.raw.title(titleText);
        return this;
    }
    /** Add a toggle checkbox button to the form. The value `key` will show
     * up as a key of {@link ModalFormResponse.prototype.formValues} with a
     * corresponding boolean value. */
    toggle(key, label, defaultValue) {
        this.#items.push(formValue => [key, formValue]);
        this.raw.toggle(label, defaultValue);
        return this;
    }
}
//# sourceMappingURL=ui.js.map
