/** Unicode-aware strings: unlike the standard String type, this one treats
 * strings as an array of Unicode code points, not UTF-8 code units.
 */
export class UniString {
    #_str;
    #_cps;
    constructor(arg) {
        if (typeof arg === "string") {
            this.#_str = arg;
        }
        else {
            this.#_cps = Array.isArray(arg) ? new Uint32Array(arg) : arg;
        }
    }
    /// Empty UniString.
    // QuickJS doesn't allow this: it follows the ES2020 semantics so
    // "UniString" will be put in the context at the very end of class
    // initialisation. "new UniString()" is not possible in property
    // initialisers.
    //public static readonly empty = new UniString("");
    //
    // So this is a workaround:
    static empty;
    get #str() {
        if (this.#_str === undefined) {
            this.#_str = String.fromCodePoint(...this.#_cps);
        }
        return this.#_str;
    }
    get #cps() {
        if (!this.#_cps) {
            this.#_cps = new Uint32Array(Array.from(this.#_str).map(c => c.codePointAt(0)));
        }
        return this.#_cps;
    }
    get length() {
        return this.#cps.length;
    }
    *[Symbol.iterator]() {
        for (const cp of this.#cps) {
            yield String.fromCodePoint(cp);
        }
    }
    at(idx) {
        const cp = this.#cps[idx];
        return cp === undefined ? undefined : String.fromCodePoint(cp);
    }
    concat(...strs) {
        let len = this.length;
        for (const str of strs) {
            len += str.length;
        }
        const cps = new Uint32Array(len);
        let pos = this.length;
        cps.set(this.#cps);
        for (const str of strs) {
            cps.set(str.#cps, pos);
            pos += str.length;
        }
        return new UniString(cps);
    }
    equals(str) {
        return this.#str === str.#str;
    }
    substring(start, end) {
        return new UniString(this.#cps.subarray(start, end));
    }
    valueOf() {
        return this.#str;
    }
}
// @ts-ignore: See above
UniString.empty = new UniString("");
//# sourceMappingURL=unicode.js.map
