var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
import { LootTable, LootCondition, LootEntry, LootPool, randomIntInClosedInterval } from "../loot-table.js";
/** Mixin for blocks that should be considered equivalent ignoring a
 * certain block state. This mixin overrides `isEquivalentTo`.
 */
export function IgnoringState(base, state) {
    class IgnoringState extends base {
        isEquivalentTo(pa, pb) {
            if (pa.typeId === pb.typeId) {
                for (const [key, value] of pa.states) {
                    if (key === state)
                        continue;
                    else if (pb.states.get(key) !== value)
                        return false;
                }
                return true;
            }
            else {
                return false;
            }
        }
    }
    return IgnoringState;
}
/** Mixin for blocks that should be considered equivalent ignoring any
 * block states. This mixin overrides `isEquivalentTo`.
 */
export function IgnoringAllStates(base) {
    class IgnoringAllStates extends base {
        isEquivalentTo(pa, pb) {
            return pa.typeId === pb.typeId;
        }
    }
    return IgnoringAllStates;
}
/// Mixin for blocks requiring a silk-touch tool to quick-mine.
export function SilkTouchForQuickMining(base) {
    class SilkTouchForQuickMining extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.enchantments.has("silk_touch"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return SilkTouchForQuickMining;
}
/** Mixin for blocks that drop something other than the mined blocks
 * themselves.
 */
export function Drops(base, loot) {
    var _Drops_loots;
    class Drops extends base {
        constructor() {
            super(...arguments);
            _Drops_loots.set(this, new LootTable().always([
                loot instanceof LootPool
                    ? loot
                    : new LootPool().entry(loot)
            ]));
        }
        lootTable() {
            return __classPrivateFieldGet(this, _Drops_loots, "f");
        }
    }
    _Drops_loots = new WeakMap();
    return Drops;
}
/** Mixin for blocks requiring a silk-touch tool to drop itself, and drop
 * nothing otherwise.
 */
export function SilkTouchForDrop(base) {
    return Fragile(base, new LootPool());
}
/** Combination of `SilkTouchForQuickMining` and `SilkTouchForDrop`. */
export function SilkTouchRequired(base) {
    return SilkTouchForQuickMining(SilkTouchForDrop(base));
}
/** Mixin for blocks that uses a different loot table when mined with a
 * non-silk-touch tool. This is a generalisation of {@link
 * SilkTouchForDrop}.
 */
export function Fragile(base, degraded) {
    class Fragile extends base {
        lootTable(perm) {
            const stack = perm.getItemStack(1);
            if (stack)
                return new LootTable()
                    .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(stack)]) // 100% drop
                    .otherwise([degraded instanceof LootPool
                        ? degraded
                        : new LootPool().entry(degraded)]); // 100% drop
            else
                // No corresponding items exist for this block. Drop nothing.
                return new LootTable();
        }
    }
    return Fragile;
}
/// Mixin for blocks that yield experience when mined with a non-silk-touch tool.
export function YieldsExperience(base, min, max = min) {
    class YieldsExperience extends base {
        experience(_perm, tool) {
            if (tool.enchantments.has("silk_touch"))
                return 0;
            else
                return randomIntInClosedInterval(min, max);
        }
    }
    return YieldsExperience;
}
/// Mixin for blocks whose drops are affected multiplicatively by Fortune.
export function MultiplicativeDrops(base, block, minRolls, maxRolls, drop) {
    const loots = new LootTable()
        .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(block)]) // 100% drop
        .otherwise([new LootPool()
            .entry(LootEntry
            .item(drop)
            .amount(minRolls, maxRolls)
            .multiplicative())
    ]);
    class MultiplicativeDrops extends base {
        lootTable(_perm) {
            return loots;
        }
    }
    return MultiplicativeDrops;
}
/// Mixin for blocks whose drops use discrete uniform distribution.
export function DiscreteUniformDrops(base, block, minRolls, maxRolls, limit, drop) {
    const loots = new LootTable()
        .when(LootCondition.matchTool().enchantment("silk_touch"), [new LootPool().entry(block)]) // 100% drop
        .otherwise([new LootPool()
            .entry(LootEntry
            .item(drop)
            .amount(minRolls, maxRolls)
            .discreteUniform(limit))
    ]);
    class DiscreteUniformDrops extends base {
        lootTable(_perm) {
            return loots;
        }
    }
    return DiscreteUniformDrops;
}
/// Mixin for blocks requiring stone-tier tools to mine.
export function StoneTier(base) {
    class StoneTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:stone_tier") ||
                tool.tags.has("minecraft:iron_tier") ||
                tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return StoneTier;
}
/// Mixin for blocks requiring iron-tier tools to mine.
export function IronTier(base) {
    class IronTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:iron_tier") ||
                tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return IronTier;
}
/// Mixin for blocks requiring diamond-tier tools to mine.
export function DiamondTier(base) {
    class DiamondTier extends base {
        isToolSuitable(perm, tool, prefs) {
            if (tool.tags.has("minecraft:diamond_tier") ||
                tool.tags.has("minecraft:netherite_tier"))
                return super.isToolSuitable(perm, tool, prefs);
            else
                return false;
        }
    }
    return DiamondTier;
}
//# sourceMappingURL=mixins.js.map
